import pysam
import subprocess
from itertools import combinations
from collections import defaultdict
import pandas as pd


class ReadOverlapGroup:
    def __init__(self, read, span=0):
        # genomic position of the read group
        self.chrom = read.reference_name
        self.start = read.reference_start
        self.end = read.reference_end

        # extend region
        self.start -= span
        self.end += span
        self.span = span
        # span apply to both cur and new read, so span 1000 means 2000 distance

        self.reads = [read]

    def add_if_overlap(self, read):
        if read.reference_name != self.chrom:
            # not the same chrom
            return False

        # apply span to new read
        start = read.reference_start
        end = read.reference_end
        start -= self.span
        end += self.span

        if (end > self.start) and (start < self.end):
            self.reads.append(read)
            self.start = min(self.start, start)
            self.end = max(self.end, end)
            # overlap
            return True
        else:
            # not overlap
            return False


class ReadSplitOverlapGroup:
    def __init__(self, read):
        # genomic position of the read group
        self.start = read.get_tag('SS')
        self.end = read.get_tag('SE')
        self.reads = [read]
        self.is_read1 = read.is_read1

    def add_if_overlap(self, read):
        if read.is_read1 != self.is_read1:
            # not the same read type
            return False

        # apply span to new read
        start = read.get_tag('SS')
        end = read.get_tag('SE')

        if (end > self.start) and (start < self.end):
            self.reads.append(read)
            self.start = min(self.start, start)
            self.end = max(self.end, end)
            # overlap
            return True
        else:
            # not overlap
            return False


def process_bam_read_pairs(read_parts):
    read_groups = []
    for read in read_parts:
        added = False
        for rg in read_groups:
            # added will be True if read overlap with existing rg
            added = rg.add_if_overlap(read)
            if added:
                break
        if not added:
            # no overlap to existing rg, create a new one
            read_groups.append(ReadSplitOverlapGroup(read))

    # select the longest alen read from each group
    final_reads = []
    for group in read_groups:
        longest_read = sorted(
            group.reads,
            key=lambda r: r.get_tag('SE') - r.get_tag('SS'),
            reverse=True)[0]
        final_reads.append(longest_read)
    return final_reads


def two_read_contact_type(read_1, read_2):
    is_same_read = read_1.is_read1 == read_2.is_read1
    if is_same_read:
        is_cutsite = (read_1.get_tag('SS') == read_2.get_tag('SE')) or \
                     (read_1.get_tag('SE') == read_2.get_tag('SS'))
    else:
        is_cutsite = False

    cis = read_1.reference_name == read_2.reference_name

    if is_same_read:
        if is_cutsite:
            return 'ciscut' if cis else 'transcut'
        else:
            return 'chimeric'
    else:
        return 'cis' if cis else 'trans'


def extract_contact_info(reads, span=1000):
    # merge reads with a span, ignore strand or read type difference
    read_groups = []
    for read in reads:
        added = False
        for rg in read_groups:
            # added will be True if read overlap with existing rg
            added = rg.add_if_overlap(read)
            if added:
                break
        if not added:
            # no overlap to existing rg, create a new one
            read_groups.append(
                ReadOverlapGroup(read,
                                 span=span))

    n_groups = len(read_groups)
    if n_groups < 2:
        # not a contact (after consider span)
        return [(None, 'no')]
    elif n_groups >= 2:
        results = []
        multi = '_multi' if n_groups > 2 else ''
        for g1, g2 in combinations(read_groups, 2):
            group_contacts = {}
            for read_1 in g1.reads:
                for read_2 in g2.reads:
                    contact_type = two_read_contact_type(read_1, read_2)
                    group_contacts[contact_type] = (read_1, read_2)
            if 'ciscut' in group_contacts:
                results += [(group_contacts['ciscut'], 'ciscut' + multi)]
            elif 'transcut' in group_contacts:
                results += [(group_contacts['transcut'], 'transcut' + multi)]
            elif 'chimeric' in group_contacts:
                results += [(None, 'chimeric')]
            elif 'cis' in group_contacts:
                results += [(group_contacts['cis'], 'cis' + multi)]
            elif 'trans' in group_contacts:
                results += [(group_contacts['trans'], 'trans' + multi)]
        return results


def _read_to_string_contact(read):
    read_type = 1 if read.is_read1 else 2
    return f'{read.reference_name}\t{read.pos}\t{read.aend}\t' \
           f'{int(read.is_reverse)}\t{read.get_tag("SS")}\t{read.get_tag("SE")}\t{read_type}'


class ContactWriter:
    def __init__(self, output_prefix):
        self.out = open(f'{output_prefix}.raw_contacts.tsv', 'w')
        title = 'read_pair_id\tcontact_type\t' \
                'chrom1\tstart1\tend1\tstrand1\tread_start1\tread_end1\tread_type1\t' \
                'chrom2\tstart2\tend2\tstrand2\tread_start2\tread_end2\tread_type2\n'
        self.out.write(title)
        self.counter = defaultdict(int)
        self.cur_read_pair = 0

    def dump(self, results):
        for result in results:
            read_pair, contact_type = result
            self.counter[contact_type] += 1
            if read_pair is None:
                continue
            read_1, read_2 = read_pair
            read_1_str = _read_to_string_contact(read_1)
            read_2_str = _read_to_string_contact(read_2)
            contact_row = f"{self.cur_read_pair}\t{contact_type}\t{read_1_str}\t{read_2_str}\n"
            self.out.write(contact_row)
        self.cur_read_pair += 1
        return

    def close(self):
        self.out.close()


def _dedup_chrom_df(chrom_df):
    dup_judge = chrom_df[['start1', 'start2', 'end1', 'end2']].apply(lambda i: i.duplicated())
    # if reads on both sides have one postion (start or end) being the same,
    # it is consider as a duplciated contact
    dup_call = (dup_judge['start1'] | dup_judge['end1']) & (dup_judge['start2'] | dup_judge['end2'])
    chrom_df = chrom_df[~dup_call]
    return chrom_df


def _dedup_contacts(output_prefix):
    input_path = f'{output_prefix}.raw_contacts.tsv'
    output_path = f'{output_prefix}.dedup_contacts.tsv.gz'
    contacts = pd.read_csv(input_path, sep='\t')
    contacts = contacts.sort_values(
        by=['chrom1', 'chrom2', 'start1', 'start2', 'end1', 'end2'])
    total_dedup = []
    for _, chrom_df in contacts.groupby(['chrom1', 'chrom2']):
        total_dedup.append(_dedup_chrom_df(chrom_df))
    total_dedup = pd.concat(total_dedup)
    total_dedup.to_csv(output_path, sep='\t', index=False)
    subprocess.run(f'rm -f {input_path}', shell=True)
    return


def _contact_to_hic_format(output_prefix):
    contact_df = pd.read_csv(f'{output_prefix}.dedup_contacts.tsv.gz', sep='\t')
    # some columns used in hic format
    contact_df['hic0'] = 0
    contact_df['hic1'] = 1
    output_path = f'{output_prefix}.3C.contact.tsv.gz'
    contact_df[[
        'strand1', 'chrom1', 'start1', 'hic0', 'strand2', 'chrom2', 'start2',
        'hic1'
    ]].to_csv(output_path, header=None, index=None, sep='\t')
    return


def process_3c_bam(bam_path, bam_prefix, contact_prefix):
    bam = pysam.AlignmentFile(bam_path)

    out_bam = pysam.AlignmentFile(f'{bam_prefix}.non_overlap_read_parts.bam',
                                  header=bam.header,
                                  mode='w')
    out_contacts = ContactWriter(contact_prefix)
    count = 0
    cur_read_pair_name = None
    cur_read_parts = []
    for read in bam:
        read_pair_name, others = read.qname.split('_')
        read_type, *_, start, stop = others.split(':')
        read.is_read1 = read_type == '1'
        read.is_read2 = not read.is_read1
        read.set_tag('SS', int(start))
        read.set_tag('SE', int(stop))
        if read_pair_name == cur_read_pair_name:
            cur_read_parts.append(read)
        else:
            # read to a new read pair
            if len(cur_read_parts) > 0:
                # process the previous read pair
                count += 1
                final_reads = process_bam_read_pairs(cur_read_parts)
                results = extract_contact_info(final_reads, span=1000)
                out_contacts.dump(results)
                for final_read in final_reads:
                    # need to put back the flag
                    # otherwise picard will raise errors
                    # dedup still perform at single read level
                    # unless the flag can be properly set
                    final_read.is_read1 = False
                    final_read.is_read2 = False
                    out_bam.write(final_read)
            # initiate the next pair
            cur_read_pair_name = read_pair_name
            cur_read_parts = [read]
    if len(cur_read_parts) > 0:
        # process the last read pair
        final_reads = process_bam_read_pairs(cur_read_parts)
        results = extract_contact_info(final_reads, span=1000)
        out_contacts.dump(results)
        for final_read in final_reads:
            # need to put back the flag
            # otherwise picard will raise errors
            # dedup still perform at single read level
            # unless the flag can be properly set
            final_read.is_read1 = False
            final_read.is_read2 = False
            out_bam.write(final_read)
    out_bam.close()
    out_contacts.close()

    # dedup contacts
    _dedup_contacts(contact_prefix)
    # save hic format
    _contact_to_hic_format(contact_prefix)

    # save counts
    stat = out_contacts.counter
    stat['mapped_frag'] = count
    pd.Series(stat).to_csv(f'{contact_prefix}.3c_bam_stats.csv', header=False)
    return



import pathlib
import pandas as pd
from .prepare_impute import execute_command


def prepare_dataset_commands(output_dir, fasta_path, cpu=10):
    output_dir = pathlib.Path(output_dir)
    project_name = output_dir.name
    scool_dir = output_dir / 'scool'
    snakemake_dir = scool_dir / 'snakemake'
    snakemake_dir.mkdir(exist_ok=True, parents=True)
    raw_dir = scool_dir / 'raw'
    raw_dir.mkdir(exist_ok=True)
    impute_dir = scool_dir / 'impute'
    impute_dir.mkdir(exist_ok=True)
    dataset_dir = scool_dir / 'dataset'
    dataset_dir.mkdir(exist_ok=True)

    # Calculate compartment at 100Kb resolution
    compartment_input_dir = impute_dir / '100K'
    compartment_cell_table = pd.Series({
        path.name.split('.')[0]: str(path)
        for path in compartment_input_dir.glob('*/*.cool')
    })
    compartment_cell_table_path = compartment_input_dir / 'cell_table.tsv'
    compartment_cell_table.to_csv(compartment_cell_table_path, sep='\t', header=None)
    # prepare a whole genome CpG ratio profile
    cpg_path = compartment_input_dir / 'cpg_ratio.hdf'
    cpg_ratio_cmd = f'hicluster cpg-ratio --cell_url {compartment_cell_table.iloc[0]} ' \
                    f'--fasta_path {fasta_path} --hdf_output_path {cpg_path}'
    execute_command(cpg_ratio_cmd)
    # compartment command
    compartment_cmd = f'hicluster compartment ' \
                      f'--cell_table_path {compartment_cell_table_path} ' \
                      f'--output_prefix {dataset_dir / project_name} ' \
                      f'--cpg_profile_path {cpg_path} ' \
                      f'--cpu {cpu}'

    # Calculate domain at 25Kb resolution
    domain_input_dir = impute_dir / '25K'
    domain_cell_table = pd.Series({
        path.name.split('.')[0]: str(path)
        for path in domain_input_dir.glob('*/*.cool')
    })
    domain_cell_table_path = domain_input_dir / 'cell_table.tsv'
    domain_cell_table.to_csv(domain_cell_table_path, sep='\t', header=None)
    domain_cmd = f'hicluster domain ' \
                 f'--cell_table_path {domain_cell_table_path} ' \
                 f'--output_prefix {dataset_dir / project_name} ' \
                 f'--resolution 25000 ' \
                 f'--window_size 10 ' \
                 f'--cpu {cpu}'

    # Calculate cell embedding/decomposition at 100Kb resolution
    embedding_dir = dataset_dir / 'embedding'
    embedding_dir.mkdir(exist_ok=True)
    embedding_cmd = f'hicluster embedding ' \
                    f'--cell_table_path {compartment_cell_table_path} ' \
                    f'--output_dir {embedding_dir} ' \
                    f'--dim 50 ' \
                    f'--dist 1000000 ' \
                    f'--resolution 100000 ' \
                    f'--scale_factor 100000 ' \
                    f'--norm_sig --save_raw ' \
                    f'--cpu {cpu}'

    # prepare qsub
    qsub_dir = snakemake_dir / 'qsub'
    qsub_dir.mkdir(exist_ok=True)
    with open(qsub_dir / 'dataset_cmd.txt', 'w') as f:
        f.write('\n'.join([compartment_cmd, domain_cmd, embedding_cmd]))
    qsub_str = f"""
#!/bin/bash
#$ -N y{project_name}
#$ -V
#$ -l h_rt=999:99:99
#$ -l s_rt=999:99:99
#$ -wd {qsub_dir}
#$ -e {qsub_dir}/qsub_dataset.error.log
#$ -o {qsub_dir}/qsub_dataset.output.log
#$ -pe smp 1
#$ -l h_vmem=3G

yap qsub --command_file_path {qsub_dir}/dataset_cmd.txt \
--working_dir {qsub_dir} --project_name y{project_name}_dataset \
--total_cpu {int(cpu*3)} --qsub_global_parms "-pe smp={cpu};-l h_vmem=5G"
"""
    with open(qsub_dir / 'qsub_dataset.sh', 'w') as f:
        f.write(qsub_str)
    return



import pandas as pd
import pathlib
import yaml
from cooler.util import binnify, read_chromsizes
import numpy as np
import time
from statsmodels.stats.multitest import multipletests
from scipy.stats import norm
import subprocess
import pybedtools
from concurrent.futures import as_completed, ProcessPoolExecutor

import cemba_data

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])
DSS_MULTI_GROUP_TEMPLATE = PACKAGE_DIR / 'dmr/dss/DSS.MultiGroup.SingleRegionDML.ipynb'


def prepare_snakemake(allc_table_path, output_dir, chrom_sizes_path, template_path, chroms=None, smoothing=True,
                      chunk_size=50000000):
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    allc_table = pd.read_csv(allc_table_path, sep='\t')
    allc_table.columns = ['allc_path', 'sample', 'group']

    chrom_sizes = read_chromsizes(chrom_sizes_path).reindex(chroms)
    if chroms is None:
        chroms = chrom_sizes.index.tolist()
    bins = binnify(chrom_sizes.loc[chroms], binsize=chunk_size)
    regions = []
    for _, (chrom, start, end) in bins.iterrows():
        region = f'{chrom}:{start}-{end}'
        regions.append(region)

    for region in regions:
        config_path = f'{output_dir}/{region}.yaml'
        parameters = dict(region=region,
                          allc_table_path=allc_table_path,
                          smoothing=smoothing)
        with open(config_path, 'w') as f:
            f.write(yaml.dump(parameters))

    snakefile = f"""
regions = {regions}
rule main:
    input:
        expand('{{region}}.DSS.DML.hdf', region=regions)

rule papermill:
    input:
        nb='{template_path}',
        config='{{region}}.yaml'
    output:
        nb='{{region}}.ipynb',
        data='{{region}}.DSS.DML.hdf'
    shell:
        'papermill {{input.nb}} {{output.nb}} -f {{input.config}} && sleep 10'
"""
    snakefile_path = f'{output_dir}/Snakefile'
    with open(snakefile_path, 'w') as f:
        f.write(snakefile)
    return snakefile_path


def _parse_dml_ids(string):
    ids = np.array(string.split(',')).astype(int)
    start = min(ids)
    end = max(ids)
    result = pd.Series({
        'idx_start': start,
        'idx_end': end,
        'n_sig': ids.size,
        'total_dml': end - start + 1,
        'sig_ratio': ids.size / (end - start + 1)
    })
    return result


def call_dmr_single_chromosome(output_dir, chrom, p_threshold, min_cg, min_len,
                               sig_ratio, chrom_sizes_path):
    # read DML for one chromosome
    print(f'Reading DML tables for {chrom}')
    dss_paths = list(pathlib.Path(output_dir).glob(f'{chrom}:*DSS.DML.hdf'))
    dmls = []
    for path in dss_paths:
        try:
            df = pd.read_hdf(path)
            if df.shape[0] > 0:
                dmls.append(df)
        except ValueError:
            # this happens when the HDF5 is empty
            pass
    dmls = pd.concat(dmls)
    dmls.sort_values('pos', inplace=True)
    dmls.dropna(subset=['pvals'], inplace=True)
    dmls.reset_index(drop=True, inplace=True)

    print('Selecting significant DMLs to merge DMRs')
    # Select sig
    # recalculate FDR
    _, fdr, *_ = multipletests(dmls['pvals'], method='fdr_bh')
    dmls['fdrs'] = fdr
    # select sig DML to merge DMR
    dmls['sig'] = dmls['fdrs'] < p_threshold
    dmls.to_hdf(f'{output_dir}/{chrom}.DML.hdf', key='data', format="table")
    sig_dmls = dmls[dmls['sig']]

    # Merge DMR
    print('Merge DMRs')
    dml_bed = sig_dmls[['chr', 'pos', 'pos']].reset_index()
    dml_bed.columns = ['id', 'chr', 'start', 'end']
    dml_bed['end'] += 1
    dml_bed = dml_bed.iloc[:, [1, 2, 3, 0]]
    dml_bed = pybedtools.BedTool.from_dataframe(dml_bed)
    try:
        dmr_bed = dml_bed.sort(g=chrom_sizes_path).merge(
            d=250, c='4', o='collapse').to_dataframe()
    except pybedtools.helpers.BEDToolsError:
        print('No DMR')
        return False

    # Annotate DMR
    print('Annotating DMRs')
    name = dmr_bed.pop('name')
    dmr_bed = pd.concat([dmr_bed, name.apply(_parse_dml_ids)], axis=1)
    dmr_bed = dmr_bed.astype({
        'idx_start': int,
        'idx_end': int,
        'n_sig': int,
        'total_dml': int
    })

    def _region_stat(row):
        idx_start = row['idx_start']
        idx_end = row['idx_end']
        return dmls.iloc[idx_start:idx_end + 1].agg({
            'stat': 'sum'
        })

    dmr_stats = dmr_bed.apply(_region_stat, axis=1)
    dmr_bed = pd.concat([dmr_bed, dmr_stats], axis=1)
    dmr_bed['length'] = dmr_bed['end'] - dmr_bed['start']

    # final DMR filter
    judge = (dmr_bed['n_sig'] >= min_cg) & (
            dmr_bed['sig_ratio'] > sig_ratio) & (dmr_bed['length'] >= min_len)
    dmr_bed['selected_dmr'] = judge
    dmr_bed.to_hdf(f'{output_dir}/{chrom}.DMR.hdf', key='data')
    return True


def run_dss_multi_group(allc_table_path, output_dir, study_name, chrom_sizes_path, chroms=None, smoothing=True,
                        p_threshold=0.001, min_cg=1, min_len=1, sig_ratio=0.5, cpu=10, save_dml=False,
                        template_path='default', chunk_size=50000000):
    # prepare template
    if template_path == 'default':
        template_path = DSS_MULTI_GROUP_TEMPLATE

    # prepare snakemake
    output_dir = pathlib.Path(output_dir).absolute()
    this_study_dir = output_dir / f'{study_name}_DSS'
    prepare_snakemake(allc_table_path=allc_table_path,
                      output_dir=this_study_dir,
                      chrom_sizes_path=chrom_sizes_path,
                      chroms=chroms,
                      smoothing=smoothing,
                      template_path=template_path,
                      chunk_size=chunk_size)
    # the ipykernel package raise "zmq.error.ZMQError: Address already in use" due to parallel execution,
    # restart likely solve the problem.
    snakemake_cmd = f'snakemake -d {this_study_dir} --snakefile {this_study_dir}/Snakefile ' \
                    f'-j {cpu} --default-resources mem_mb=100 --resources mem_mb={int(5000 * cpu)} ' \
                    f'--restart-times 3'
    subprocess.run(snakemake_cmd, shell=True, check=True)

    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for chrom in chroms:
            f = exe.submit(call_dmr_single_chromosome,
                           output_dir=this_study_dir,
                           chrom=chrom,
                           p_threshold=p_threshold,
                           min_cg=min_cg,
                           min_len=min_len,
                           sig_ratio=sig_ratio,
                           chrom_sizes_path=chrom_sizes_path)
            futures[f] = chrom
            time.sleep(1)

        total_dmrs = []
        for future in as_completed(futures):
            chrom = futures[future]
            print(f'{chrom} finished')
            success_flag = future.result()
            if success_flag:
                total_dmrs.append(pd.read_hdf(f'{this_study_dir}/{chrom}.DMR.hdf'))
        if len(total_dmrs) != 0:
            total_dmrs = pd.concat(total_dmrs)
            total_dmrs.to_hdf(f'{this_study_dir}/{study_name}_DMR.hdf', key='data')

    for chrom in chroms:
        subprocess.run(f'rm -f {this_study_dir}/{chrom}:*.DSS.DML.hdf', shell=True)
        subprocess.run(f'rm -f {this_study_dir}/{chrom}:*.yaml', shell=True)
        subprocess.run(f'rm -f {this_study_dir}/{chrom}.DMR.hdf', shell=True)
        if not save_dml:
            subprocess.run(f'rm -f {this_study_dir}/{chrom}.DML.hdf', shell=True)
        subprocess.run(f'rm -rf {this_study_dir}/.snakemake {this_study_dir}/Snakefile', shell=True)
        subprocess.run(f'rm -f {this_study_dir}/*log {this_study_dir}/qsub.sh', shell=True)
    return snakemake_cmd



"""
Contain codes about parse plate info and generate sample sheet
"""

import pathlib
import re
from collections import OrderedDict

import pandas as pd

import cemba_data

# Load defaults
PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])

# the Illumina sample sheet header used by Ecker Lab
with open(PACKAGE_DIR / 'files/sample_sheet_header.txt') as _f:
    SAMPLESHEET_DEFAULT_HEADER = _f.read()

SECTIONS = ['[CriticalInfo]', '[LibraryInfo]', '[PlateInfo]']

LIMITED_CHOICES = {
    'n_random_index': [8, 384, '8', '384'],
    'input_plate_size': [384, '384'],
    'primer_quarter': ['Set1_Q1', 'Set1_Q2', 'Set1_Q3', 'Set1_Q4',
                       'SetB_Q1', 'SetB_Q2', 'SetB_Q3', 'SetB_Q4']}

CRITICAL_INFO_KEYS = ['n_random_index', 'input_plate_size',
                      'pool_id', 'tube_label', 'email']

# key (n_random_index, input_plate_size)
BARCODE_TABLE = {
    ('8', '384'): PACKAGE_DIR / 'files/V1_i7_i5_index.tsv',  # V1 can use both Set1 and SetB i5 i7 primer
    ('384', '384'): PACKAGE_DIR / 'files/V2_i7_i5_index.tsv'  # V2 only use SetB primer
}


def _clean_str_for_path(str_in):
    # replace special char with _
    str_out = re.sub('[^a-zA-Z0-9]', '_', str_in.strip())
    return str_out


def _get_kv_pair(line):
    try:
        k, v = line.split('=')
        if k == 'email':
            return k, v
        else:
            return _clean_str_for_path(k), _clean_str_for_path(v)
    except ValueError:
        raise ValueError(f'Each key=value line must contain a "=" to separate key and value. Got {line}')


def _read_plate_info(plate_info_path):
    """Parse the plate info file"""
    cur_section = ''
    cur_section_id = -1

    critical_info = {}
    library_info = OrderedDict()
    plate_header = True
    plate_info = []

    with open(plate_info_path) as f:
        for line in f:
            line = line.strip('\n')
            if line == '' or line.startswith('#'):
                continue

            # determine section
            if line.startswith('['):
                cur_section_id += 1
                if line == SECTIONS[cur_section_id]:
                    cur_section = line
                else:
                    raise ValueError(
                        f'Section name and order must be [CriticalInfo] [LibraryInfo] [PlateInfo], '
                        f'got {line} at No.{cur_section_id + 1} section.')
            elif cur_section == '[CriticalInfo]':
                k, v = _get_kv_pair(line)
                if k not in CRITICAL_INFO_KEYS:
                    raise ValueError(f'Unknown key {k} in [CriticalInfo]')
                else:
                    critical_info[k] = v
            elif cur_section == '[LibraryInfo]':
                k, v = _get_kv_pair(line)
                if (k in critical_info.keys()) or (k in library_info.keys()):
                    raise ValueError(f'Found duplicated key {k}')
                else:
                    library_info[k] = v
            elif cur_section == '[PlateInfo]':
                # support more general characters
                ll = re.split(r' |,|\t', 'line')
                # ll = line.split('\t')
                if plate_header:
                    plate_header = False
                plate_info.append(ll)
            else:
                raise ValueError(f'Got a malformed line: {line}')

    for k in CRITICAL_INFO_KEYS:
        if k not in critical_info:
            raise ValueError(f'[CriticalInfo] missing key-value pair "{k}"')

    header = plate_info[0]
    plate_info = pd.DataFrame(plate_info[1:], columns=plate_info[0])
    for k, v in library_info.items():
        if k in plate_info.columns:
            raise ValueError(f'Found duplicated key {k} between [PlateInfo] and [LibraryInfo]')
        plate_info[k] = v
    if critical_info['n_random_index'] == '8':
        n_plate_info_fix_col = 2
        if plate_info['plate_id'].duplicated().sum() != 0:
            raise ValueError(f'Found duplicated plate_id in [PlateInfo] section.')
    elif critical_info['n_random_index'] == '384':
        n_plate_info_fix_col = 3
        if plate_info.set_index(['plate_id', 'multiplex_group']).index.duplicated().sum() != 0:
            raise ValueError(f'Found duplicated plate_id, multiplex_group combination in [PlateInfo] section.')
    else:
        raise ValueError(f'[CriticalInfo] n_random_index got unknown value '
                         f'{critical_info["n_random_index"]}')
    col_order = header[:n_plate_info_fix_col] + list(library_info.keys()) + header[n_plate_info_fix_col:]
    plate_info = plate_info[col_order].copy()
    plate_info['sample_id_prefix'] = plate_info.apply(
        lambda i: '-'.join(i[n_plate_info_fix_col:].astype(str).tolist()), axis=1)

    # after getting sample_id_prefix, add critical info into plate_info too
    for k, v in critical_info.items():
        if k in plate_info.columns:
            raise ValueError(f'Found duplicated key {k}')
        plate_info[k] = v

    # deduplicate primer name
    # because we start to use the same primer for different multiplex groups
    plate_info = plate_info.loc[~plate_info['primer_name'].duplicated(), :].copy()
    return critical_info, plate_info


def reverse_comp(sequence):
    rc_dict = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C',
               'a': 't', 't': 'a', 'c': 'G', 'g': 'C',
               'n': 'n', 'N': 'N'}
    _seq = ''.join([rc_dict[s] for s in sequence[::-1]])
    return _seq


def _plate_384_random_index_8(plate_info, barcode_table, i5_reverse_comp=False):
    """UID pattern of V1 {sample_id_prefix}-{plate1}-{plate2}-{plate_pos}"""
    records = []

    # check plate_info primer compatibility
    for primer_quarter, n_plate in plate_info['primer_quarter'].value_counts().iteritems():
        if n_plate < 2:
            print(f'{primer_quarter} only have 1 plate in the table, please make sure this is correct.')
        elif n_plate == 2:
            pass
        else:
            raise ValueError(f'{primer_quarter} have {n_plate} plates in the table, that is impossible.')

    for primer_quarter, plate_pair in plate_info.groupby('primer_quarter'):
        if primer_quarter not in LIMITED_CHOICES['primer_quarter']:
            raise ValueError(f'Unknown primer_quarter value {primer_quarter}')

        if plate_pair.shape[0] == 1:
            # sometimes, a quarter only index 1 plate
            plate1, *plate2 = plate_pair['plate_id']
            plate2 = 'GHOST_PLATE'
        else:
            plate1, plate2 = plate_pair['plate_id']

        # check plate pair info consistence
        for col_name, col in plate_pair.iteritems():
            if col.unique().size != 1:
                if col_name != 'plate_id':
                    print(f'{col_name} contains different information between {plate1} and {plate2}, '
                          f'Will put {plate1} prefix into sample_id. This should not happen normally.')

        # remove all the special char with '_' in plate names
        # I use '-' to separate sample parts
        plate1 = _clean_str_for_path(plate1)
        plate2 = _clean_str_for_path(plate2)

        for col in 'ABCDEFGH':
            for row in range(1, 13):
                plate_pos = f'{col}{row}'
                cur_row = barcode_table.loc[(primer_quarter, plate_pos)]
                i5_barcode = cur_row['i5_index_sequence']
                i7_barcode = cur_row['i7_index_sequence']
                sample_id_prefix = plate_pair['sample_id_prefix'].iloc[0]
                sample_id = f'{sample_id_prefix}-{plate1}-{plate2}-{plate_pos}'

                # THIS IS BASED ON FORMAT BCL2FASTQ NEEDS
                records.append({'Sample_ID': sample_id,
                                'index': i7_barcode,  # the index must be i7
                                'index2': reverse_comp(i5_barcode) if i5_reverse_comp else i5_barcode,
                                # the index2 must be i5
                                'Sample_Project': plate_pair['tube_label'].iloc[0],
                                'Description': plate_pair['email'].iloc[0]})

    miseq_sample_sheet, nova_sample_sheet = _make_final_samplesheet(records)
    return miseq_sample_sheet, nova_sample_sheet


def _plate_384_random_index_384(plate_info, barcode_table, i5_reverse_comp=False):
    """
    UID pattern of V2 {sample_id_prefix}-{plate}-{multiplex_group}-{barcode_name}

    If i5_reverse_comp, use the reverse complement i5 sequence, this is needed for NovaSeq v1.5 S4 kit.
    """
    records = []

    # this is now possible because we may used the same PCR index for the same plate
    # # check plate_info primer compatibility
    # for primer_name, n_primer in plate_info['primer_name'].value_counts().iteritems():
    #     if n_primer > 1:
    #         raise ValueError(f'{primer_name} have {n_primer} multiplex_group in the table, that is impossible.')

    for _, row in plate_info.iterrows():
        plate = row['plate_id']
        # remove all the '-' with '_' in plate names
        plate = _clean_str_for_path(plate)

        barcode_name = row['primer_name']
        cur_row = barcode_table.loc[barcode_name]
        i5_barcode = cur_row['i5_index_sequence']
        i7_barcode = cur_row['i7_index_sequence']
        sample_id_prefix = row['sample_id_prefix']
        multiplex_group = row['multiplex_group']
        sample_id = f'{sample_id_prefix}-{plate}-{multiplex_group}-{barcode_name}'

        # THIS IS BASED ON FORMAT BCL2FASTQ NEEDS
        records.append({'Sample_ID': sample_id,
                        'index': i7_barcode,  # the index must be i7
                        'index2': reverse_comp(i5_barcode) if i5_reverse_comp else i5_barcode,  # the index2 must be i5
                        'Sample_Project': row['tube_label'],
                        'Description': row['email']})

    miseq_sample_sheet, nova_sample_sheet = _make_final_samplesheet(records)
    return miseq_sample_sheet, nova_sample_sheet


def _make_final_samplesheet(records):
    # THIS IS BASED ON FORMAT BCL2FASTQ NEEDS
    sample_sheet = pd.DataFrame(records)
    sample_sheet['Sample_Name'] = ''
    sample_sheet['Sample_Well'] = ''
    sample_sheet['I7_Index_ID'] = ''
    sample_sheet['I5_Index_ID'] = ''
    sample_sheet['I7_Index_ID'] = ''
    sample_sheet['Sample_Plate'] = 'Plate'

    miseq_sample_sheet = sample_sheet[['Sample_ID', 'Sample_Name', 'Sample_Plate',
                                       'Sample_Well', 'I7_Index_ID', 'index',
                                       'I5_Index_ID', 'index2', 'Sample_Project',
                                       'Description']].copy()

    lanes = []
    for i in range(1, 5):
        lane_df = miseq_sample_sheet.copy()
        lane_df['Lane'] = i
        lanes.append(lane_df)
    nova_sample_sheet = pd.concat(lanes)
    nova_sample_sheet = nova_sample_sheet[['Lane', 'Sample_ID', 'Sample_Name', 'Sample_Plate',
                                           'Sample_Well', 'I7_Index_ID', 'index',
                                           'I5_Index_ID', 'index2', 'Sample_Project',
                                           'Description']].copy()
    return miseq_sample_sheet, nova_sample_sheet


def make_sample_sheet(plate_info_path: str, output_prefix: str, header_path=None):
    """
    make two sample sheets for bcl2fastq based on the plate info file: one for miseq, one for novaseq

    Parameters
    ----------
    plate_info_path
    output_prefix
    header_path

    Returns
    -------

    """
    # read plate info
    critical_info, plate_info = _read_plate_info(plate_info_path)

    # check valid choice
    for k in ['n_random_index', 'input_plate_size']:
        if critical_info[k] not in LIMITED_CHOICES[k]:
            raise ValueError(f'Invalid value in critical_info section for {k}: {critical_info[k]}')

    n_random_index = critical_info['n_random_index']
    input_plate_size = critical_info['input_plate_size']
    barcode_table_path = BARCODE_TABLE[n_random_index, input_plate_size]
    for i5_reverse_comp in [True, False]:
        if (n_random_index, input_plate_size) == ('8', '384'):
            barcode_table = pd.read_csv(barcode_table_path, sep='\t')
            barcode_table['primer_quarter'] = barcode_table['Index_set'] + "_" + barcode_table['Index_quarter']
            barcode_table.set_index(['primer_quarter', 'plate_pos'], inplace=True)
            miseq_sample_sheet, nova_sample_sheet = _plate_384_random_index_8(plate_info, barcode_table,
                                                                              i5_reverse_comp=i5_reverse_comp)
        elif (critical_info['n_random_index'], critical_info['input_plate_size']) == ('384', '384'):
            barcode_table = pd.read_csv(barcode_table_path,
                                        sep='\t', index_col='set_384_plate_pos')
            miseq_sample_sheet, nova_sample_sheet = _plate_384_random_index_384(plate_info, barcode_table,
                                                                                i5_reverse_comp=i5_reverse_comp)
        else:
            raise NotImplementedError(f"Unknown combination of n_random_index {n_random_index} "
                                      f"and input_plate_size {input_plate_size}")

        # before write out, check plate info compatibility:
        # check plate_info primer compatibility
        if int(n_random_index) == 8:
            for primer_quarter, n_plate in plate_info['primer_quarter'].value_counts().iteritems():
                if n_plate < 2:
                    print(f'{primer_quarter} only have 1 plate, please make sure this is right.')
                elif n_plate == 2:
                    pass
                else:
                    raise ValueError(f'{primer_quarter} have {n_plate} plates in the table, that is impossible.')
        elif int(n_random_index) == 384:
            pass
            # this is now possible, we use the same index for one plate
            # for primer_name, n_primer in plate_info['primer_name'].value_counts().iteritems():
            #     if n_primer > 1:
            #         raise ValueError(f'{primer_name} have {n_primer} multiplex_group in the table, '
            #                          f'that is impossible.')
        else:
            raise ValueError(f'Unknown n_random_index {n_random_index}.')

        # write miseq sample sheet
        if not i5_reverse_comp:
            with open(output_prefix + '.miseq.sample_sheet.csv', 'w') as output_f:
                if header_path is None:
                    output_f.write(SAMPLESHEET_DEFAULT_HEADER)
                else:
                    with open(header_path) as hf:
                        output_f.write(hf.read())
                output_f.write(miseq_sample_sheet.to_csv(index=None))

        # write novaseq sample sheet
        with open(output_prefix + f'.novaseq.{"V1.5" if i5_reverse_comp else "v1"}.sample_sheet.csv', 'w') as output_f:
            if header_path is None:
                output_f.write(SAMPLESHEET_DEFAULT_HEADER)
            else:
                with open(header_path) as hf:
                    output_f.write(hf.read())
            output_f.write(nova_sample_sheet.to_csv(index=None))
    return


def print_plate_info(barcode_version):
    if barcode_version.upper() == 'V1':
        with open(PACKAGE_DIR / 'files/plate_info_template_v1.txt') as f:
            template = f.read()
    elif barcode_version.upper() == 'V2':
        with open(PACKAGE_DIR / 'files/plate_info_template_v2.txt') as f:
            template = f.read()
    else:
        raise ValueError(f'Primer Version can only be V1 or V2, got {barcode_version}.')
    print(template)
    return



"""
SGE qsub system auto-submitter
"""

import datetime
import json
import os
import re
import shutil
import sys
import time
from subprocess import run, PIPE, CalledProcessError

import pandas as pd

SUBMISSION_GAP = 0.1
QSTAT_GAP_INIT = 5
REQUIRED_KEYS = ['pe smp', 'l h_vmem']


def _default_command_dict(name, error_path, output_path, working_dir,
                          h_rt='999:99:99', s_rt='999:99:99'):
    """Default qsub command dict"""
    command_dict = {'command': None,
                    'N': name,
                    'V': '',
                    'l h_rt': h_rt,
                    'l s_rt': s_rt,
                    'wd': working_dir,
                    'e': error_path,
                    'o': output_path}
    return command_dict


def _get_running_job_id_qstat(user_name, id_set):
    """
    Check qstat of user_name, get running job_ids to determine job status

    Parameters
    ----------
    user_name
    id_set

    Returns
    -------
    return active ids in the submitted id_set;
    or return all active ids if id_set is None
    """

    while True:
        tried = 0
        try:
            qstat_result_string = run(['qstat', '-u', user_name],
                                      stdout=PIPE, encoding='utf8', check=True).stdout
            break
        except CalledProcessError as e:
            # there is a timeout error from qstat occasionally during some mapping (when system in high load?)
            # "ERROR: failed receiving gdi request response for mid=1 (got syncron message receive timeout error)"
            # current solution is re-try for 10 times, otherwise quit the job.
            # This timeout error usually do not happen again in retry.
            tried += 1
            time.sleep(30)
            if tried > 10:
                print(f'Qstat failed {tried} times continuously, '
                      f'seems that the qsub system have some problem')
                raise e
    if 'job' not in qstat_result_string:
        # SGE specific?
        # qstat print nothing, all job finished
        return []
    else:
        qstat_result_lines = qstat_result_string.split('\n')
        total_id = []
        for line in qstat_result_lines:
            # Typical 1st line of qstat
            # job-ID prior name user state submit/start at queue slots ja-task-ID
            line_id = line.strip().split(' ')[0]  # id part
            if id_set is not None:
                # only record id within id_set (submitted jobs by this project)
                if line_id in id_set:
                    total_id.append(line_id)
            else:
                # record any id if id_set is not provided
                total_id.append(line_id)
        return total_id


class _Qsubmitter:
    """
    _Qsubmitter take single command_file_path, execute all the commands in this file in parallel using qsub.
    """

    def __init__(self, command_file_path, working_dir, project_name, force_redo=False, global_parm_dict=None,
                 total_cpu=60, total_mem=500, submission_gap=SUBMISSION_GAP, qstat_gap=QSTAT_GAP_INIT,
                 resubmit_failed=True, merge_jobs=None):
        # prepare working_dir
        self.working_dir = str(working_dir)
        if project_name[0].isdigit():
            raise ValueError('Project name can not start with number, qsub will fail.')
        self.project_name = project_name
        self.project_dir = self.working_dir + '/' + self.project_name + '_qsub'
        if os.path.exists(self.project_dir):
            if force_redo:
                shutil.rmtree(self.project_dir)
                os.mkdir(self.project_dir)
            else:
                print('Note: The project has been submitted before. Finished jobs will not run again.')
                print(f'If you want to redo everything, set force_redo=True')
                sys.stdout.flush()
        else:
            os.mkdir(self.project_dir)
        run(['cp', command_file_path, self.project_dir])

        # submission parameters
        self.total_cpu = total_cpu
        self.total_mem = total_mem  # GBs

        self.submission_gap = submission_gap
        self.qstat_gap = qstat_gap

        # whether resubmit failed jobs
        self.resubmit_failed = resubmit_failed

        # auto detect user name
        self.user_name = run(['whoami'], stdout=PIPE, encoding='utf8').stdout.strip()
        print(f'Username: {self.user_name}')

        # prepare project dir
        self.command_file_path = command_file_path
        self.global_parm_dict = global_parm_dict if global_parm_dict is not None else {}
        if not os.path.exists(self.project_dir):
            os.mkdir(self.project_dir)

        # get all Command objects
        # self.commands is a list of command objects
        self.commands = self._parse_command_file(merge_jobs=merge_jobs)

        # initiate running stats
        self.running_commands = []
        self.running_cpu = 0
        self.running_mem = 0  # GBs
        self.submitted_qsub_id = set()
        self.submission_fail_commands = []
        self.finished_commands = []
        self.finish_count = 0
        self.success_commands = []
        self.success_count = 0

        # add a signal file in project dir
        self.alive = True
        self.core_signal_path = self.project_dir + '/RUNNING_SIGNAL.json'
        with open(self.core_signal_path, 'w') as f:
            core_signal_dict = {'total_cpu': self.total_cpu,
                                'total_mem': self.total_mem,
                                'alive': self.alive}
            self.core_signal_dict = core_signal_dict
            json.dump(core_signal_dict, f)
        print(f"The CORE SIGNAL json file is generated: {self.core_signal_path}.\n"
              f"You can modify value (not key) in this json file "
              f"to change hot-coding parameters of the queue.\n"
              f"IMPORTANT: If 'alive' in this file is not true, qsub will stop submit "
              f"new jobs and exit until all current jobs finished.")

        # submit jobs
        self.submit()

        # get job reports
        self.get_reports()
        return

    def _parse_command_file(self, merge_jobs=None):
        with open(self.command_file_path) as command_f:
            if self.command_file_path.endswith('json'):
                command_dict_list = json.load(command_f)
            else:
                print('\ncommand file name does not end with "json", '
                      'will read each line in the file as a single command.\n'
                      'In this case, you can not specify qsub parameters in the command file '
                      'for each job separately, you can only pass qsub parameter through '
                      '--qsub_global_parms, this will apply to every command in this command file.\n'
                      'If you want to specify qsub parameters for each job separately, '
                      'use JSON format command file.\n')
                total_commands = command_f.readlines()

                command_dict_list = []
                if (merge_jobs is None) or (merge_jobs >= len(total_commands)):
                    for line in total_commands:
                        command_dict_list.append({'command': line})
                else:
                    print(f'Will merge commands into {merge_jobs} qsub jobs.')
                    job_lists = [[] for _ in range(merge_jobs)]
                    for i, line in enumerate(total_commands):
                        job_lists[i % merge_jobs].append(line)
                    for single_job_list in job_lists:
                        # in this case, the command value would be a list, will detect in writing command file
                        command_dict_list.append({'command': single_job_list})

            # validate command file
            if not isinstance(command_dict_list, list):
                raise ValueError(f'json.load(open(command_file)) does not return a list, '
                                 f'but got {type(command_dict_list)}. '
                                 f'Command file path: {self.command_file_path}')
            if len(command_dict_list) < 1:
                raise ValueError(f'Command file seems empty...'
                                 f'Command file path: {self.command_file_path}')

            obj_list = []
            for n, command_dict in enumerate(command_dict_list):
                command_dict.update(self.global_parm_dict)
                obj_list.append(_Command(command_dict=command_dict,
                                         unique_id=f'{self.project_name}_{n}',
                                         working_dir=self.working_dir,
                                         project_dir=self.project_dir))
        return obj_list

    def submit(self):
        # job submission process:
        for command_obj in self.commands:
            if command_obj.finish:
                # happens when command_obj.status_path exist
                self.finished_commands.append(command_obj)
                if self.resubmit_failed:
                    # resubmit_failed commands, and only skip success command
                    if command_obj.success:
                        self.success_commands.append(command_obj)
                        continue
                else:
                    # skip all the finish command (has .json file), no matter fail or success
                    continue
            command_cpu = int(command_obj.qsub_parameter['pe smp'])
            # here is the total memory, need to multiply by command_cpu
            command_mem = command_cpu * int(command_obj.qsub_parameter['l h_vmem'].strip('G'))
            future_cpu = self.running_cpu + command_cpu
            future_mem = self.running_mem + command_mem
            # check both cpu and mem
            if (future_cpu > self.total_cpu) or (future_mem > self.total_mem):
                print(f'Pause submitting jobs, '
                      f'current running CPU {self.running_cpu} / {self.total_cpu}; '
                      f'current running MEM {self.running_mem} / {self.total_mem}.')
                # submitted enough jobs, wait until running_cpu or running_mem decrease
                temp_gap = self.qstat_gap
                while True:
                    # after check_running, self.running_cpu and self.running_mem will be updated.
                    # self.total_cpu, self.total_mem, self.alive may also change, if detected RUNNING_SIGNAL change
                    self.check_running()
                    future_cpu = self.running_cpu + command_cpu
                    future_mem = self.running_mem + command_mem
                    if (future_cpu <= self.total_cpu) and (future_mem <= self.total_mem):
                        break
                    time.sleep(temp_gap)
                    # wait longer and longer if the job is stuck
                    if temp_gap < 300:
                        temp_gap += 5
            if not self.alive:
                print(f'Stop submitting jobs because CORE SIGNAL "alive" is not true. '
                      f'CORE SIGNAL dict path {self.core_signal_path}')
                break
            command_obj.submit()
            # gather submitted id and obj
            if command_obj.submission_fail:
                self.submission_fail_commands.append(command_obj)
            else:
                self.running_commands.append(command_obj)
                self.submitted_qsub_id.add(command_obj.qsub_id)
            self.running_cpu += command_cpu
            self.running_mem += command_mem
            time.sleep(self.submission_gap)

        # final job check:
        temp_gap = self.qstat_gap
        while True:
            self.check_running()
            if self.running_cpu == 0:
                break  # all job finished
            time.sleep(temp_gap)
            if temp_gap < 60:
                temp_gap += 5
        return

    def check_running(self):
        # In this function, check
        # 1. running jobs, their info and resource consumption
        # 2. core_signal_dict, to adjust global control
        print('check running job: ', end='')
        cur_running_qsub_id = _get_running_job_id_qstat(user_name=self.user_name,
                                                        id_set=self.submitted_qsub_id)

        # Part 1: check every running obj
        cur_running_cpu = 0
        cur_running_mem = 0  # GBs
        cur_running_command = []
        for command_obj in self.running_commands:
            if command_obj.qsub_id not in cur_running_qsub_id:
                # print(self.submitted_qsub_id)
                # print(cur_running_qsub_id)
                # print(command_obj.qsub_id, command_obj.unique_id, 'finishing')

                # command have finished in this qsubmitter
                # need to manually set finish to True, and read log to generate status json again
                # if old status exist, will be overwrote
                command_obj.finish = True
                command_obj.check_output_log()
                self.finish_count += 1
                self.finished_commands.append(command_obj)
                self.submitted_qsub_id.remove(command_obj.qsub_id)

                # check if the finished job is success
                if command_obj.success:
                    self.success_count += 1
                    self.success_commands.append(command_obj)
            else:
                cur_running_command.append(command_obj)
                command_cpu = int(command_obj.qsub_parameter['pe smp'])
                command_mem = int(command_cpu * int(command_obj.qsub_parameter['l h_vmem'].strip('G')))
                cur_running_cpu += command_cpu
                cur_running_mem += command_mem

        print(f'{self.finish_count} job finished in this submission. {self.success_count} job success.')
        sys.stdout.flush()

        self.running_commands = cur_running_command
        # update running cpu
        self.running_cpu = cur_running_cpu
        self.running_mem = cur_running_mem

        # part 2: core_signal_dict
        try:
            with open(self.core_signal_path) as f:
                self.core_signal_dict = json.load(f)
                if self.total_mem != self.core_signal_dict['total_mem']:
                    print(f"Change the queue total MEM to {self.core_signal_dict['total_mem']}")
                    self.total_mem = self.core_signal_dict['total_mem']
                if self.total_cpu != self.core_signal_dict['total_cpu']:
                    print(f"Change the queue total CPU to {self.core_signal_dict['total_cpu']}")
                    self.total_cpu = self.core_signal_dict['total_cpu']
                self.alive = self.core_signal_dict['alive']
        except OSError:
            print(f'Warning: Unable to load CORE SIGNAL json file, '
                  f'will not update core signal dict. '
                  f'This autogenerated file is a dict json, '
                  f'should exist in {self.command_file_path}.')
        return

    def get_reports(self):
        command_records = []
        for command in self.commands:
            if os.path.exists(command.status_path):
                status_series = pd.Series(command.check_submitted_status())
            else:
                status_series = pd.Series({})
            status_series['unique_id'] = command.unique_id
            status_series['submission_fail'] = command.submission_fail
            command_records.append(status_series)

        status_df = pd.DataFrame(command_records)
        status_df.set_index('unique_id', inplace=True)
        status_df.to_csv(self.project_dir + '/summary_status.tsv', sep='\t')

        print('Job submission status:')
        failed = status_df['submission_fail'].sum()
        print(f'\t{failed} / {status_df.shape[0]} job(s) failed in submission.')

        print('Job exit status:')
        for code, sub_df in status_df[~status_df['submission_fail']].groupby('return_code'):
            running_time = sub_df.duration_second.mean()
            time_str = str(datetime.timedelta(seconds=running_time))
            print(f'\t{sub_df.shape[0]} job(s) end with return code {code:.0f}, '
                  f'the average running time is {time_str}')
        return


class _Command:
    """Command object for single command script file"""

    def __init__(self, command_dict, unique_id, working_dir, project_dir):
        self.project_dir = project_dir
        self.unique_id = unique_id  # unique id is given by submitter, qsub_id is given by qsub system

        # command status
        self.submitted_in_this_run = False
        self.qsub_id = None
        self.submission_fail = False
        self.finish = False  # if submitted is True and not in qstat. For resubmit, if self.status_path exist.
        self.success = False  # if return code == 0
        self.status_path = f'{self.project_dir}/{unique_id}.json'
        self.script_path = f'{self.project_dir}/{unique_id}.sh'
        self.error_path = f'{self.project_dir}/{unique_id}.error.log'
        self.output_path = f'{self.project_dir}/{unique_id}.output.log'
        self.start_time = None  # will be datetime.datetime after check_output_log
        self.end_time = None  # will be datetime.datetime after check_output_log
        self.duration_second = .0
        self.return_code = -99

        # prepare command dict
        self.command_dict = _default_command_dict(name=self.unique_id,
                                                  error_path=self.error_path,
                                                  output_path=self.output_path,
                                                  working_dir=working_dir)
        # print('default', self.command_dict)
        self.command_dict.update(**command_dict)
        # print('from input', command_dict)
        # print('after update', self.command_dict)
        # sys.stdout.flush()
        for key_name in REQUIRED_KEYS:
            if key_name not in self.command_dict:
                raise KeyError(f'Required Key {key_name} not found in command_dict')

        for k, v in self.command_dict.items():
            if v is None:
                raise ValueError(f'{k} is None in command dict')

        # separate command and qsub parameter
        self.command = self.command_dict.pop('command')
        if isinstance(self.command, str):
            self.command = self.command.split(';')
        # from here self.command must be a list
        self.n_command = len(self.command)

        self.qsub_parameter = self.command_dict
        self.command_dict = None

        # make command sh file
        self._make_script_file()
        # if self.status_path exist, self.submitted, self.finish will be true
        self.check_submitted_status()
        return

    def _make_script_file(self):
        """Make command file, add some echos to print start, end and return code"""
        if os.path.exists(self.script_path):
            return

        with open(self.script_path, 'w') as sh:
            sh.write("#!/bin/bash\n")
            for k, v in self.qsub_parameter.items():
                if k.startswith('l '):
                    sh.write(f'#$ -{k}={v}\n')
                else:
                    sh.write(f'#$ -{k} {v}\n')

            # self.command must be a list
            for command_id, command in enumerate(self.command):
                sh.write(f'echo JOB_CMD_START {self.unique_id} COMMAND {command_id} $(date +"%H:%M:%S-%D")\n')
                sh.write(command + '\n')
                # check command return code
                sh.write(f'echo JOB_CMD_RETURN_CODE {self.unique_id} COMMAND {command_id} $?\n')
                sh.write(f'echo JOB_CMD_END {self.unique_id} COMMAND {command_id} $(date +"%H:%M:%S-%D")\n')
        return

    def submit(self):
        """Submit the qsub script"""
        job_id_pattern = re.compile(r'(?<=Your job )\d+')
        return_obj = run(['qsub', self.script_path], stdout=PIPE, encoding='utf8')
        print(f'Submitted {self.unique_id}, stdout: {return_obj.stdout.strip()}')
        try:
            self.qsub_id = job_id_pattern.search(return_obj.stdout).group()
        except AttributeError:
            self.submission_fail = True
        self.submitted_in_this_run = True
        return

    def check_submitted_status(self):
        if not os.path.exists(self.status_path):
            # do noting
            return None
        try:
            with open(self.status_path) as f:
                status_dict = json.load(f)
        except json.decoder.JSONDecodeError:
            print('Fail to load status dict, treat as failed.')
            return None

        self.finish = True
        self.qsub_id = status_dict['qsub_id']
        self.start_time = datetime.datetime.strptime(status_dict['start_time'],
                                                     "%H:%M:%S-%m/%d/%y")
        self.end_time = datetime.datetime.strptime(status_dict['end_time'],
                                                   "%H:%M:%S-%m/%d/%y")
        self.duration_second = status_dict['duration_second']
        self.return_code = status_dict['return_code']
        try:
            self.success = int(self.return_code) == 0
        except ValueError:
            self.success = False
        return status_dict

    def write_status(self):
        # store the qsub id and job status.
        with open(self.status_path, 'w') as f:
            json.dump({'qsub_id': self.qsub_id,
                       'start_time': self.start_time.strftime(format='%H:%M:%S-%m/%d/%y'),
                       'end_time': self.end_time.strftime(format='%H:%M:%S-%m/%d/%y'),
                       'duration_second': self.duration_second,
                       'return_code': self.return_code}, f)
        return

    def check_output_log(self):
        """
        Each job may have multiple commands, however the job running time only consider the first start and last end
        also the return code is 0 only if all commands has return code 0, otherwise -99
        So everything in status file is still job level.
        """
        if not self.finish:
            # job not finish, do nothing
            return
        try:
            first_time = True
            temp_return_code_list = []
            with open(self.output_path) as f:
                for line in f:
                    if (self.unique_id in line) and line.startswith('JOB_CMD'):
                        ll = line.strip().split(' ')
                        if line.startswith('JOB_CMD_START') and first_time:
                            # only record the first command start time
                            self.start_time = datetime.datetime.strptime(ll[-1], "%H:%M:%S-%m/%d/%y")
                            first_time = False
                        elif line.startswith('JOB_CMD_END'):
                            # and record the last command end time
                            self.end_time = datetime.datetime.strptime(ll[-1], "%H:%M:%S-%m/%d/%y")
                        elif line.startswith('JOB_CMD_RETURN_CODE'):
                            try:
                                temp_return_code_list.append(int(ll[-1]))
                            except ValueError:
                                temp_return_code_list.append(-99)
                        else:
                            continue
            # judge the overall return code if all commands, if saw any non-zero code, will set to -99, otherwise 0
            self.return_code = 0
            if len(temp_return_code_list) == 0:
                self.return_code = -99
            for return_code in temp_return_code_list:
                if return_code != 0:
                    self.return_code = -99

        except FileNotFoundError:
            self.return_code = -99
            self.start_time = None
            self.end_time = None
            print(f'Output log for job {self.qsub_id} {self.unique_id} not found, this job may failed.')

        if (self.start_time is None) or (self.end_time is None):
            print(f'Job {self.qsub_id} start time: {self.start_time} (default is None)')
            print(f'Job {self.qsub_id} end time: {self.end_time} (default is None)')
            print(f'Output log for job {self.qsub_id} {self.unique_id} is incomplete, this job may failed.')
            self.return_code = -99
            self.start_time = datetime.datetime.now()
            self.end_time = datetime.datetime.now()

        self.duration_second = (self.end_time - self.start_time).total_seconds()
        self.success = int(self.return_code) == 0

        # write status json to self.status_path
        self.write_status()
        return


def qsub(command_file_path,
         working_dir,
         project_name,
         wait_until=None,
         total_cpu=60,
         total_mem=500,
         qsub_global_parms='',
         force_redo=False,
         submission_gap=SUBMISSION_GAP,
         qstat_gap=QSTAT_GAP_INIT,
         resubmit_failed=True,
         merge_jobs=300):
    """
    TODO write doc
    TODO fix error problem

    Parameters
    ----------
    command_file_path
    working_dir
    project_name
    wait_until
    total_cpu
    total_mem
    qsub_global_parms
    force_redo
    submission_gap
    qstat_gap
    resubmit_failed
    merge_jobs

    Returns
    -------

    """
    working_dir = str(working_dir)
    # deal with qsub_global_parms:
    global_parm_dict = {}
    for parm_pair in qsub_global_parms.split(';'):
        parm_pair = parm_pair.strip().strip("'").strip('"').lstrip('-')
        if parm_pair != '':
            if parm_pair.count('=') == 0:
                global_parm_dict[parm_pair] = ''
            elif parm_pair.count('=') == 1:
                k, v = parm_pair.split('=')
                global_parm_dict[k] = v
            else:
                raise ValueError(f'Can not parse global parm part: "{parm_pair}"')
    # print(global_parm_dict)

    if isinstance(command_file_path, str):
        command_file_path = [command_file_path]

    # wait the jobs in {wait_until} to finish, then start this qsub
    if wait_until is not None:
        if isinstance(wait_until, str):
            wait_until = {wait_until}
        elif isinstance(wait_until, list):
            wait_until = set(wait_until)
        else:
            raise TypeError(f'wait_until should either be str or list, '
                            f'got {type(wait_until)}')

        snap = 60
        while True:
            user_name = run(['whoami'], stdout=PIPE, encoding='utf8').stdout.strip()
            cur_running_qsub_id = _get_running_job_id_qstat(user_name, id_set=wait_until)
            if len(cur_running_qsub_id) == 0:
                break
            print(f'Still waiting {len(cur_running_qsub_id)} jobs to finish ', end='')
            if len(cur_running_qsub_id) < 10:
                job_ids = ' '.join(cur_running_qsub_id)
                print(f'Their qsub ids are: {job_ids}')
            time.sleep(snap)
            snap += 60
            snap = min(snap, 300)

    # for multiple command files, run one by one
    for i, command_file in enumerate(command_file_path):
        print(f'Execute command file: {command_file}. '
              f'Total command file executed [{i + 1} / {len(command_file_path)}]')
        _project_name = project_name
        if len(command_file_path) != 1:
            # make sure project name is unique
            _project_name += f'_{i}'
        submitter = _Qsubmitter(command_file_path=command_file,
                                global_parm_dict=global_parm_dict,
                                working_dir=working_dir,
                                project_name=_project_name,
                                total_cpu=total_cpu,
                                total_mem=total_mem,
                                submission_gap=submission_gap,
                                qstat_gap=qstat_gap,
                                force_redo=force_redo,
                                resubmit_failed=resubmit_failed,
                                merge_jobs=merge_jobs)
        print(f'{len(submitter.commands)} jobs processed in this qsubmitter. From command file {command_file}')
    return



from .stats_parser import *


def snmc_summary():
    """
    Generate snmC pipeline MappingSummary.csv.gz and save into cwd

    Returns
    -------
    pd.DataFrame
    """
    all_stats = []

    # fastq trimming stats
    df = parse_single_stats_set(f'fastq/*.trimmed.stats.txt',
                                cell_parser_cutadapt_trim_stats)
    all_stats.append(df)

    # hisat-3n mapping
    df = parse_single_stats_set(f'bam/*.hisat3n_dna_summary.txt',
                                cell_parser_hisat_summary)
    all_stats.append(df)

    # uniquely mapped reads dedup
    df = parse_single_stats_set(f'bam/*.unique_align.deduped.matrix.txt',
                                cell_parser_picard_dedup_stat, prefix='UniqueAlign')
    all_stats.append(df)

    # multi mapped reads dedup
    df = parse_single_stats_set(f'bam/*.multi_align.deduped.matrix.txt',
                                cell_parser_picard_dedup_stat, prefix='MultiAlign')
    all_stats.append(df)

    # allc count
    df = parse_single_stats_set(f'allc/*.allc.tsv.gz.count.csv',
                                cell_parser_allc_count)
    all_stats.append(df)

    # concatenate all stats
    all_stats = pd.concat(all_stats, axis=1)
    all_stats.index.name = 'cell'
    all_stats.to_csv(f'MappingSummary.csv.gz')
    return all_stats


def snmct_summary():
    """
    Generate snmCT pipeline MappingSummary.csv.gz and save into cwd

    Returns
    -------
    pd.DataFrame
    """
    all_stats = []

    # fastq trimming stats
    df = parse_single_stats_set(f'fastq/*.trimmed.stats.txt',
                                cell_parser_cutadapt_trim_stats)
    all_stats.append(df)

    # hisat-3n DNA mapping
    df = parse_single_stats_set(f'bam/*.hisat3n_dna_summary.txt',
                                cell_parser_hisat_summary, prefix='DNA')
    all_stats.append(df)

    # hisat-3n RNA mapping
    df = parse_single_stats_set(f'rna_bam/*.hisat3n_rna_summary.txt',
                                cell_parser_hisat_summary, prefix='RNA')
    all_stats.append(df)

    # uniquely mapped reads dedup
    df = parse_single_stats_set(f'bam/*.unique_align.deduped.matrix.txt',
                                cell_parser_picard_dedup_stat, prefix='DNAUniqueAlign')
    all_stats.append(df)

    # multi mapped reads dedup
    df = parse_single_stats_set(f'bam/*.multi_align.deduped.matrix.txt',
                                cell_parser_picard_dedup_stat, prefix='DNAMultiAlign')
    all_stats.append(df)

    # uniquely mapped dna reads selection
    df = parse_single_stats_set('bam/*.hisat3n_dna.unique_align.deduped.dna_reads.reads_mch_frac.csv',
                                cell_parser_reads_mc_frac_profile, prefix='UniqueAlign')
    all_stats.append(df)

    # multi mapped dna reads selection
    df = parse_single_stats_set('bam/*.hisat3n_dna.multi_align.deduped.dna_reads.reads_mch_frac.csv',
                                cell_parser_reads_mc_frac_profile, prefix='MultiAlign')
    all_stats.append(df)

    # uniquely mapped rna reads selection
    df = parse_single_stats_set('rna_bam/*.hisat3n_rna.unique_align.rna_reads.reads_mch_frac.csv',
                                cell_parser_reads_mc_frac_profile)
    all_stats.append(df)

    # allc count
    df = parse_single_stats_set(f'allc/*.allc.tsv.gz.count.csv',
                                cell_parser_allc_count)
    all_stats.append(df)

    # feature count
    df = parse_single_stats_set(f'rna_bam/*.feature_count.tsv.summary',
                                cell_parser_feature_count_summary)
    all_stats.append(df)

    # concatenate all stats
    all_stats = pd.concat(all_stats, axis=1)
    all_stats.index.name = 'cell'
    all_stats.to_csv(f'MappingSummary.csv.gz')

    # TODO
    # 2. aggregate feature count this can be a separate func and call in snakefile
    return all_stats



import pathlib
from collections import defaultdict

import pandas as pd
import pysam


def _determine_reads_conversion(read):
    """
    Determine HISAT-3N read conversion level

    Parameters
    ----------
    read

    Returns
    -------
    return conversion string GA or CT
    """

    # For HISAT-3N, use YZ tag
    # YZ:A:<A>: The value + or – indicate the read is mapped to REF-3N (+) or REF-RC-3N (-).
    # actually, for snmC, the + means G to A conversion; the - means C to T conversion.
    yz_tag = read.get_tag('YZ')
    if yz_tag == '-':
        # G -> A
        return 'GA'
    elif yz_tag == '+':
        # C -> T
        return 'CT'
    else:
        raise ValueError


def _complement(seq):
    d = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'N': 'N'}
    try:
        return ''.join([d[b] for b in seq])
    except KeyError:
        return seq


def _determine_mch_context(conversion, ref_pos, ref_seq_dict, nome=False):
    """
    Determine whether the base is mCH context or not. If nome=True, skip GpC sites.

    Parameters
    ----------
    conversion :
        Type of conversion, GA or CT
    ref_pos :
        Position of the base in the reference
    ref_seq_dict :
        Dictionary of the reference sequence
    nome :
        If True, skip all the GpC context when calculating read mCH fraction,
        as it is subject to NOMe methylation

    Returns
    -------
    flag : bool
    """
    flag = False
    try:
        if conversion == 'GA':  # this means G to A conversion
            # the actual C is on the reverse strand
            ref_context = ref_seq_dict[ref_pos] + ref_seq_dict[ref_pos - 1]
            ref_context = _complement(ref_context)
        else:  # this means C to T conversion
            # the actual C is on the forward strand
            ref_context = ref_seq_dict[ref_pos] + ref_seq_dict[ref_pos + 1]

        if ref_context in {'CC', 'CT', 'CA'}:
            flag = True
            if nome:
                if conversion == 'GA':
                    if ref_seq_dict[ref_pos + 1] == 'G':
                        flag = False
                else:
                    if ref_seq_dict[ref_pos - 1] == 'G':
                        flag = False
    except KeyError:
        # the base is at boarder
        pass
    return flag


def _single_read_mch_level(read, nome=False, frac=False):
    """
    Determine the mCH fraction of single read

    Parameters
    ----------
    read
        Read object from bam file
    nome
        If True, skip all the GpC context as it is subject to methylation
    frac
        If True, return the mCH fraction of the read instead of mCH count

    Returns
    -------

    """
    read_conversion = _determine_reads_conversion(read)

    # ref seq is parsed based on read seq and MD tag, and do not depend on reverse or not
    ref_seq = read.get_reference_sequence().upper()
    ref_pos = read.get_reference_positions()
    # use dict instead of string is because ref_seq could contain blocks when skip happen
    ref_seq_dict = {pos: base for pos, base in zip(ref_pos, ref_seq)}
    read_seq = read.seq.upper()

    # only count mCH
    mch = 0
    cov = 0
    other_snp = 0
    # read.get_aligned_pairs always return small to large position
    for read_pos, ref_pos, ref_base in read.get_aligned_pairs(
            matches_only=True, with_seq=True):
        read_base = read_seq[read_pos]
        ref_read_pair = ref_base + read_base
        ref_read_pair = ref_read_pair.upper()
        is_mch_context = _determine_mch_context(read_conversion, ref_pos, ref_seq_dict, nome=nome)
        if is_mch_context:
            if ref_read_pair in ['CC', 'GG']:  # means unconverted and methylated
                cov += 1
                mch += 1
            elif ref_read_pair in ['CT', 'GA']:  # means converted and un-methylated
                cov += 1
            else:
                # other kinds of SNPs, do not count to cov
                other_snp += 1
                pass
    if frac:
        read_mch_frac = (mch / cov) if cov > 0 else 0
        return read_mch_frac, cov, other_snp
    else:
        return mch, cov, other_snp


def select_mct_reads_normal(input_bam,
                            output_bam,
                            mode,
                            mc_rate_max_threshold=None,
                            mc_rate_min_threshold=None,
                            cov_min_threshold=3,
                            nome=False):
    """
    Select DNA reads with mCH fraction <= mc_rate_max_threshold and
    coverage >= cov_min_threshold from mCT bam file

    Parameters
    ----------
    input_bam :
        Input bam file
    output_bam :
        Output bam file
    mode :
        Mode of reads selection, 'dna' or 'rna'
    mc_rate_max_threshold :
        Maximum mCH fraction cutoff, applied only when mode is 'dna'
    mc_rate_min_threshold :
        Minimum mCH fraction cutoff, applied only when mode is 'rna'
    cov_min_threshold :
        Minimum coverage cutoff
    nome :
        If True, skip all the GpC context when calculating read mCH fraction,
        as it is subject to NOMe methylation

    Returns
    -------

    """
    if mode == 'dna':
        if mc_rate_max_threshold is None:
            raise ValueError('mc_rate_max_threshold is required when mode is dna')
    elif mode == 'rna':
        if mc_rate_min_threshold is None:
            raise ValueError('mc_rate_min_threshold is required when mode is rna')
    else:
        raise ValueError('mode should be dna or rna')

    read_profile_dict = defaultdict(int)
    # init dict to make sure the series has something
    read_profile_dict[(50, 50)] = 0

    with pysam.AlignmentFile(input_bam) as f:
        with pysam.AlignmentFile(output_bam, header=f.header,
                                 mode='wb') as out_f:
            for read in f:
                mc_frac, cov, _ = _single_read_mch_level(read, nome=nome, frac=True)
                read_profile_dict[(int(100 * mc_frac), cov)] += 1

                # split reads
                if mode == 'dna':
                    # determine dna reads
                    if (mc_frac > mc_rate_max_threshold) or (cov < cov_min_threshold):
                        continue
                else:
                    # determine rna reads
                    if (mc_frac < mc_rate_min_threshold) or (cov < cov_min_threshold):
                        continue
                out_f.write(read)

    with open(str(output_bam)[:-4] + '.reads_mch_frac.csv', 'w') as stat_f:
        # save parameters, so the stat parser can read from it
        stat_f.write(f'#mode={mode}\n')
        stat_f.write(f'#mc_rate_max_threshold={mc_rate_max_threshold}\n')
        stat_f.write(f'#mc_rate_min_threshold={mc_rate_min_threshold}\n')
        stat_f.write(f'#cov_min_threshold={cov_min_threshold}\n')

        stat_f.write('mc_frac,cov,count\n')
        for (mc_frac, cov), count in read_profile_dict.items():
            stat_f.write(f'{mc_frac},{cov},{count}\n')
    return


def aggregate_feature_counts():
    cell_datas = []

    save_info = True
    for path in pathlib.Path('rna_bam/').glob('*.feature_count.tsv'):
        table = pd.read_csv(path, comment='#', sep='\t', index_col=0)
        cell_data = table.iloc[:, -1].squeeze()
        cell_data.name = cell_data.name.split(':')[-1]
        cell_datas.append(cell_data)
        if save_info:
            table.iloc[:, :-1].to_csv('featureCounts.gene_info.csv.gz')
            save_info = False

    pd.DataFrame(cell_datas).to_csv('featureCounts.data.csv.gz')
    return



MERGE_TEMPLATE = '''
# Example (required) parameters
# merge_allc_cpu = 10
# chrom_size_path = 'PATH_TO_CHROM_SIZE_FILE'
# merge_sample_prefixes = '[]'
# copy_sample_prefixes = '[]'
# group = 'GROUP_NAME'
sample_prefixes = merge_sample_prefixes + copy_sample_prefixes

# the main rule is the final target
rule main:
    input:
        expand("{sample}.allc.tsv.gz", sample=sample_prefixes),
        expand("{sample}.allc.tsv.gz.tbi", sample=sample_prefixes),
#    output:
#        f"{group}.finished"
#    shell:
#        "date > {output}"
        


# Merge ALLC
rule merge_allc:
    input:
        "{sample}.pathlist",
    output:
        allc="{sample}.allc.tsv.gz",
        tbi="{sample}.allc.tsv.gz.tbi"
    threads:
        max(1, min(int(1.1 * merge_allc_cpu), int(workflow.cores / 1.1)))
    resources:
        mem_mb=merge_allc_cpu * 5000
    run:
        if wildcards.sample in merge_sample_prefixes:
            shell("allcools merge-allc "
                  "--allc_paths {input} "
                  "--output_path {output.allc} "
                  "--chrom_size_path {chrom_size_path} "
                  "--cpu {threads}")
        else:
            shell("cp $(cat {input}) {output.allc} ;"
                  "cp $(cat {input}).tbi {output.tbi} ;")

'''

MERGE_EXTRACT_TEMPLATE = '''
# Example (required) parameters
# merge_allc_cpu = 10
# mcg_context = 'CGN'
# chrom_size_path = 'PATH_TO_CHROM_SIZE_FILE'
# merge_sample_prefixes = '[]'
# copy_sample_prefixes = '[]'
# group = 'GROUP_NAME'
sample_prefixes = merge_sample_prefixes + copy_sample_prefixes

# the main rule is the final target
rule main:
    input:
        expand("{sample}.{mcg_context}-Merge.allc.tsv.gz", sample=sample_prefixes, mcg_context=[mcg_context]),
        expand("{sample}.{mcg_context}-Merge.allc.tsv.gz.tbi", sample=sample_prefixes, mcg_context=[mcg_context]),
#    output:
#        f"{group}.finished"
#    shell:
#        "date > {output}"
        

# Merge ALLC
rule merge_allc:
    input:
        "{sample}.pathlist",
    output:
        allc="{sample}.allc.tsv.gz",
        tbi="{sample}.allc.tsv.gz.tbi"
    threads:
        max(1, min(int(1.1 * merge_allc_cpu), int(workflow.cores / 1.1)))
    resources:
        mem_mb=merge_allc_cpu * 5000
    run:
        if wildcards.sample in merge_sample_prefixes:
            shell("allcools merge-allc "
                  "--allc_paths {input} "
                  "--output_path {output.allc} "
                  "--chrom_size_path {chrom_size_path} "
                  "--cpu {threads}")
        else:
            shell("cp $(cat {input}) {output.allc} ;"
                  "cp $(cat {input}).tbi {output.tbi} ;")

# Extract mCG ALLC for DMR calling
rule extract_allc_mcg:
    input:
        "{sample}.allc.tsv.gz"
    output:
        allc_cg="{sample}.{mcg_context}-Merge.allc.tsv.gz",
        allc_cg_tbi="{sample}.{mcg_context}-Merge.allc.tsv.gz.tbi"
    threads:
        1
    resources:
        mem_mb=100
    shell:
        "allcools extract-allc "
        "--allc_path {input} "
        "--output_prefix {wildcards.sample} "
        "--mc_contexts {mcg_context} "
        "--chrom_size_path {chrom_size_path} "
        "--strandness merge "
        "--output_format allc "
        "--cpu {threads}"
'''



import pathlib

import pandas as pd
import pysam

from .mc import mc_mapping_stats


def _count_reads_by_rg_in_star_bam(bam_path):
    try:
        bam = pysam.AlignmentFile(bam_path)
    except ValueError:
        # empty bam file
        return

    cell_read_counts = {cell['ID']: 0 for cell in bam.header['RG']}

    for read in bam:
        cell = read.get_tag('RG')
        cell_read_counts[cell] += 1
    read_count = pd.Series(cell_read_counts, name='Reads')
    read_count.index.name = 'cell_id'
    return read_count


def summary_rna_mapping(output_dir):
    output_dir = pathlib.Path(output_dir)

    # summarize read counts for each cell before filter by mC rate
    total_star_mapped_reads = _count_reads_by_rg_in_star_bam(output_dir / 'rna_bam/TotalRNAAligned.filtered.bam')

    # feature count summary
    total_counts = pd.read_csv(output_dir / 'rna_bam/TotalRNAAligned.rna_reads.feature_count.tsv.summary',
                               sep='\t', index_col=0).T
    total_counts.index = total_counts.index.map(lambda i: i.split(':')[-1])
    feature_count_summary = total_counts[['Assigned']].copy()
    feature_count_summary['FinalRNAReads'] = total_counts.sum(axis=1)
    feature_count_summary.columns = ['FinalCountedReads', 'FinalRNAReads']

    total_rna_stat = feature_count_summary.copy()
    total_rna_stat['RNAUniqueMappedReads'] = total_star_mapped_reads
    total_rna_stat['SelectedRNAReadsRatio'] = total_rna_stat['FinalRNAReads'] / total_rna_stat['RNAUniqueMappedReads']
    total_rna_stat.index.name = 'cell_id'
    return total_rna_stat


def summarize_select_dna_reads(output_dir,
                               config):
    bam_dir = pathlib.Path(output_dir) / 'bam'
    mc_rate_max_threshold = float(config['mc_rate_max_threshold'])
    cov_min_threshold = float(config['dna_cov_min_threshold'])

    records = []
    select_dna_reads_stat_list = list(bam_dir.glob('*.reads_profile.csv'))
    for path in select_dna_reads_stat_list:
        try:
            _df = pd.read_csv(path)
        except pd.errors.EmptyDataError:
            # means the bam file is empty
            continue

        cell_id = path.name.split('.')[0]
        if cell_id.endswith('-R1') or cell_id.endswith('-R2'):
            # select DNA preformed in R1 R2 separately:
            cell_id = cell_id[:-3]
        _df['cell_id'] = cell_id
        _df['mc_rate_max_threshold'] = mc_rate_max_threshold
        _df['cov_min_threshold'] = cov_min_threshold
        records.append(_df)
    total_stats_df = pd.concat(records)

    selected_reads = total_stats_df[
        (total_stats_df['cov'] >= cov_min_threshold)
        & (total_stats_df['mc_frac'] < mc_rate_max_threshold)]

    selected_reads = selected_reads.groupby('cell_id')['count'].sum()
    selected_ratio = selected_reads / total_stats_df.groupby('cell_id')['count'].sum()
    final_stat = pd.DataFrame({'FinalDNAReads': selected_reads, 'SelectedDNAReadsRatio': selected_ratio})
    final_stat.index.name = 'cell_id'
    return final_stat


def mct_mapping_stats(output_dir, config):
    """this may apply to single UID dir, so config is provided as parameter"""
    mc_stats_df = mc_mapping_stats(output_dir, config)
    select_dna_stats_df = summarize_select_dna_reads(output_dir, config)
    rna_stats_df = summary_rna_mapping(output_dir)
    final_df = pd.concat([mc_stats_df, select_dna_stats_df, rna_stats_df], axis=1)
    return final_df


def aggregate_feature_counts(output_dir):
    output_dir = pathlib.Path(output_dir)
    cell_data = []

    count_paths = list(output_dir.glob('*/rna_bam/TotalRNAAligned.rna_reads.feature_count.tsv'))
    if len(count_paths) == 0:
        return

    data = None
    for path in count_paths:
        data = pd.read_csv(path, sep='\t', index_col=0, comment='#')
        cell_data.append(data.iloc[:, 5:])
    cell_data = pd.concat(cell_data, axis=1, sort=True)
    cell_data.columns = cell_data.columns.str.split(':').str[1]

    # all count table should have the same info, so read the last one
    # chr, start, end, strand, length
    gene_info = data.iloc[:, :5]
    with pd.HDFStore(output_dir / 'TotalRNAData.h5', mode='w', complevel=5) as hdf:
        hdf['data'] = cell_data.T  # cell by gene
        hdf['gene'] = gene_info
        hdf['stats'] = pd.DataFrame({'GenesDetected': (cell_data > 0).sum()})
    return


def mct_additional_cols(final_df, output_dir):
    final_df = final_df.copy()
    final_df['CellInputReadPairs'] = final_df['R1InputReads'].astype(int)  # == final_df['R2InputReads']
    if 'PCRIndex' in final_df.columns:  # plate info might not exist if the cell name is abnormal
        cell_barcode_ratio = pd.concat([(i['CellInputReadPairs'] / i['CellInputReadPairs'].sum())
                                        for _, i in final_df.groupby('PCRIndex')])
        final_df['CellBarcodeRatio'] = cell_barcode_ratio

    stats = pd.read_hdf(output_dir / 'TotalRNAData.h5', key='stats')
    final_df['GenesDetected'] = stats['GenesDetected']

    # calculate some mCT specific ratios
    final_df['DNAReadsYield'] = final_df['FinalDNAReads'] / (
            final_df['CellInputReadPairs'] * 2)
    final_df['RNAReadsYield'] = final_df['FinalRNAReads'] / final_df[
        'CellInputReadPairs']
    final_df['RNA/(DNA+RNA)'] = final_df['FinalRNAReads'].fillna(0) / (
            final_df['R1FinalBismarkReads'].fillna(0) + 1)
    return final_df



import configparser
import functools
import itertools
import logging
import pathlib
import subprocess
from concurrent.futures import ProcessPoolExecutor, as_completed

# logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def get_configuration(config_path):
    """
    Read .ini config file from given path
    """
    if isinstance(config_path, configparser.ConfigParser):
        return config_path
    ref_path_config = configparser.ConfigParser()
    ref_path_config.read(config_path)

    total_config = {}
    for name, section in ref_path_config.items():
        for k, v in section.items():
            total_config[k] = v
    return total_config


def test_cmd(tool_name, cmd_list):
    try:
        subprocess.run(cmd_list,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE,
                       encoding='utf8',
                       check=True)
    except subprocess.CalledProcessError as e:
        log.error(f'Test {tool_name} got non-zero return code {e.returncode}')
        log.error(e.stderr)
        raise
    return


def valid_environments(config):
    log.info('Test mapping environments')

    # test cutadapt
    test_cmd(tool_name='cutadapt', cmd_list=['cutadapt', '--version'])
    # test samtools
    test_cmd(tool_name='samtools', cmd_list=['samtools', '--version'])
    # test picard, picard always have return code 1...
    test_cmd(tool_name='picard', cmd_list=['which', 'picard'])
    # test bismark_mapping
    test_cmd(tool_name='bismark_mapping', cmd_list=['bismark_mapping', '--version'])
    if config['mode'] != 'm3c':
        # test bowtie2
        test_cmd(tool_name='bowtie2', cmd_list=['bowtie2', '--version'])
    else:
        # test bowtie1
        test_cmd(tool_name='bowtie', cmd_list=['bowtie', '--version'])
    # test pigz
    test_cmd(tool_name='pigz', cmd_list=['pigz', '-V'])

    bismark_dir = pathlib.Path(config['bismark_reference'])
    if not bismark_dir.is_dir():
        raise TypeError(f"Bismark reference must be a directory contain a sub-dir named Bisulfite_Genome, "
                        f"generated by bismark_genome_preparation. Got a file path")
    if not bismark_dir.exists():
        raise FileNotFoundError(f"Bismark reference directory not found. "
                                f"Path in the config.ini is {bismark_dir}")

    allc_ref_fasta = pathlib.Path(config['reference_fasta'])
    allc_ref_fai = pathlib.Path(config['reference_fasta'] + '.fai')
    if not allc_ref_fasta.exists():
        raise FileNotFoundError(f"Reference fasta for ALLC generation not found. "
                                f"Path in the config.ini is {allc_ref_fasta}")
    if not allc_ref_fai.exists():
        raise FileNotFoundError(f".fai index for reference fasta not found. "
                                f"Path of fadix should be {allc_ref_fai}. "
                                f"You can use 'samtools fadix {allc_ref_fasta}' to generate.")
    return


def parse_index_fasta(fasta_path):
    records = {}
    with open(fasta_path) as f:
        key_line = True
        for line in f:
            if key_line:
                key = line.lstrip('>').rstrip('\n')
                key_line = False
            else:
                value = line.lstrip('^').rstrip('\n')
                records[key] = value
                key_line = True
    return records


def command_runner(commands, runner=None, cpu=1):
    if runner is None:
        from functools import partial
        runner = partial(subprocess.run,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         encoding='utf8',
                         shell=True,
                         check=True)
    if cpu <= 1:
        for command in commands:
            runner(command)
    else:
        with ProcessPoolExecutor(cpu) as pool:
            futures = []
            for command in commands:
                future = pool.submit(runner, command)
                futures.append(future)

            for future in as_completed(futures):
                try:
                    future.result()
                except subprocess.CalledProcessError as e:
                    print("Got error in fastq_qc, command was:")
                    print(command)
                    print(e.stdout)
                    print(e.stderr)
                    raise e
    return


def snakemake(workdir, snakefile, cores):
    try:
        subprocess.run([
            'snakemake', '-d', str(workdir), '--snakefile',
            str(snakefile), '--cores',
            str(cores)
        ],
            check=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            encoding='utf8')
    except subprocess.CalledProcessError as e:
        print(e.stdout)
        print(e.stderr)
        raise e
    return


def get_barcode_version(output_dir):
    fastq_dir = pathlib.Path(output_dir) / 'fastq'
    with open(fastq_dir / '.barcode_version') as f:
        return f.read()


def get_mode(output_dir):
    fastq_dir = pathlib.Path(output_dir) / 'fastq'
    with open(fastq_dir / '.mode') as f:
        return f.read()


MAPPING_MODE_CHOICES = ['mct', 'mc', 'm3c', '4m']
IUPAC_TABLE = {
    'A': 'A',
    'T': 'T',
    'C': 'C',
    'G': 'G',
    'R': 'AG',
    'Y': 'CT',
    'S': 'GC',
    'W': 'AT',
    'K': 'GT',
    'M': 'AC',
    'B': 'CGT',
    'D': 'AGT',
    'H': 'ATC',
    'V': 'ACG',
    'N': 'ATCGN'
}


@functools.lru_cache(maxsize=100)
def parse_mc_pattern(pattern: str) -> set:
    """
    parse mC context pattern
    """
    # IUPAC DNA abbr. table
    all_pos_list = []
    pattern = pattern.upper()
    for base in pattern:
        try:
            all_pos_list.append(IUPAC_TABLE[base])
        except KeyError:
            raise KeyError(f'Base {base} is not in IUPAC table.')
    context_set = set([''.join(i) for i in itertools.product(*all_pos_list)])
    return context_set



from collections import defaultdict
import re
import pandas as pd
import pysam

REVERSE_READ_MCH_CONTEXT = {'CA', 'CC', 'CT'}
FORWARD_READ_MCH_CONTEXT = {'AG', 'TG', 'GG'}


def single_read_mch_level(read, nome=False, frac=False):
    """

    Parameters
    ----------
    read
    nome
        If True, skip all the GpC context as it is subject to methylation
    frac

    Returns
    -------

    """
    # ref seq is parsed based on read seq and MD tag, and do not depend on reverse or not
    ref_seq = read.get_reference_sequence().upper()
    ref_pos = read.get_reference_positions()
    # use dict instead of string is because ref_seq could contain blocks when skip happen
    ref_seq_dict = {pos: base for pos, base in zip(ref_pos, ref_seq)}
    read_seq = read.seq.upper()

    # only count mCH
    mch = 0
    cov = 0
    other_snp = 0
    if read.is_reverse:  # read in reverse strand
        for read_pos, ref_pos, ref_base in read.get_aligned_pairs(
                matches_only=True, with_seq=True):
            read_base = read_seq[read_pos]
            ref_read_pair = ref_base + read_base
            try:
                ref_context = ref_seq_dict[ref_pos] + ref_seq_dict[ref_pos + 1]
                if nome:
                    if ref_seq_dict[ref_pos - 1] == 'G':
                        continue
                if ref_context not in REVERSE_READ_MCH_CONTEXT:
                    continue
            except KeyError:
                # ref_seq_dict KeyError means position is on border or not continuous, skip that
                continue
            if ref_read_pair == 'CC':  # C to C means unconverted and methylated
                cov += 1
                mch += 1
            elif ref_read_pair == 'CT':  # C to T means converted and un-methylated
                cov += 1
            else:
                # other kinds of SNPs, do not count to cov
                other_snp += 1
                pass
    else:  # read in forward strand
        for read_pos, ref_pos, ref_base in read.get_aligned_pairs(
                matches_only=True, with_seq=True):
            read_base = read_seq[read_pos]
            ref_read_pair = ref_base + read_base
            try:
                ref_context = ref_seq_dict[ref_pos - 1] + ref_seq_dict[ref_pos]
                if nome:
                    if ref_seq_dict[ref_pos + 1] == 'C':
                        continue
                if ref_context not in FORWARD_READ_MCH_CONTEXT:
                    continue
            except KeyError:
                # ref_seq_dict KeyError means position is on border or not continuous, skip that
                continue
            if ref_read_pair == 'GG':  # G to G means unconverted and methylated
                cov += 1
                mch += 1
            elif ref_read_pair == 'GA':  # G to A means converted and un-methylated
                cov += 1
            else:
                # other kinds of SNPs, do not count to cov
                other_snp += 1
                pass
    if frac:
        return mch, cov, other_snp
    else:
        read_mch_frac = (mch / cov) if cov > 0 else 0
        return read_mch_frac, cov, other_snp


def select_rna_reads_normal(input_bam,
                            output_bam,
                            mc_rate_min_threshold=0.9,
                            cov_min_threshold=3,
                            nome=False):
    read_profile_dict = defaultdict(int)
    with pysam.AlignmentFile(input_bam) as bam:
        with pysam.AlignmentFile(output_bam, header=bam.header,
                                 mode='wb') as out_bam:
            for read in bam:
                read_mch_rate, cov, other_snp = single_read_mch_level(read, nome=nome)
                read_profile_dict[(int(100 * read_mch_rate), cov)] += 1

                # split reads
                if (read_mch_rate <
                    mc_rate_min_threshold) or (cov < cov_min_threshold):
                    continue
                out_bam.write(read)

    with open(str(output_bam) + '.reads_profile.csv', 'w') as stat_f:
        stat_f.write('mc_frac,cov,count\n')
        for (mc_rate, cov), count in read_profile_dict.items():
            stat_f.write(f'{mc_rate},{cov},{count}\n')
    return


def select_rna_reads_split_reads(input_bam,
                                 output_bam,
                                 mc_rate_min_threshold=0.9,
                                 cov_min_threshold=3,
                                 nome=False):
    splited_read_name_pattern = re.compile('.+-[lrm]$')

    # first pass: determine read methylation level
    read_level_mcs = defaultdict(int)
    read_level_covs = defaultdict(int)
    with pysam.AlignmentFile(input_bam) as f:
        for read in f:
            mc, cov, other_snp = single_read_mch_level(read, frac=True, nome=nome)
            read_name = read.qname
            if splited_read_name_pattern.search(read_name):
                read_level_mcs[read_name[:-2]] += mc
                read_level_covs[read_name[:-2]] += cov
            else:
                read_level_mcs[read_name] += mc
                read_level_covs[read_name] += cov
    read_level_data = pd.DataFrame({
        'mc': read_level_mcs,
        'cov': read_level_covs
    })
    read_level_data['mc_frac'] = read_level_data['mc'] / (read_level_data['cov'] +
                                                       0.001)
    read_level_data['mc_frac'] = (read_level_data['mc_frac'] * 100).astype(int)
    profile = read_level_data.groupby('mc_frac')['cov'].value_counts()
    profile.name = 'count'
    profile = profile.reset_index()
    profile.to_csv(f'{output_bam}.reads_profile.csv', index=None)

    # filter reads
    use_reads = read_level_data[
        (read_level_data['mc_frac'] > mc_rate_min_threshold)
        & (read_level_data['cov'] >= cov_min_threshold)].index.tolist()
    use_reads = set(use_reads)
    del read_level_data

    # second pass: write passed reads
    with pysam.AlignmentFile(input_bam) as f:
        with pysam.AlignmentFile(output_bam, header=f.header,
                                 mode='wb') as out_f:
            for read in f:
                read_name = read.qname
                if (read_name in use_reads) or (read_name[:-2] in use_reads):
                    # read name or read name without suffix
                    out_f.write(read)
    return


def select_rna_reads(input_bam,
                     output_bam,
                     mc_rate_min_threshold=0.5,
                     cov_min_threshold=3,
                     nome=False,
                     assay_type='mc'):
    if assay_type == 'mc':
        select_rna_reads_normal(input_bam,
                                output_bam,
                                mc_rate_min_threshold=mc_rate_min_threshold,
                                cov_min_threshold=cov_min_threshold,
                                nome=nome)
    elif assay_type == 'm3c':
        select_rna_reads_split_reads(input_bam,
                                     output_bam,
                                     mc_rate_min_threshold=mc_rate_min_threshold,
                                     cov_min_threshold=cov_min_threshold,
                                     nome=nome)
    else:
        raise ValueError(f'Unknown assay_type {assay_type}.')
    return



import pathlib

import cemba_data
from ..utilities import MAPPING_MODE_CHOICES

# Load defaults
PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])


def print_default_mapping_config(mode,
                                 barcode_version,
                                 genome_fasta,
                                 bismark_ref=None,
                                 hisat3n_dna_ref=None,
                                 hisat3n_rna_ref=None,
                                 star_ref=None,
                                 gtf=None,
                                 nome=False,
                                 chrom_size_path=None):
    mode = mode.lower()
    if mode not in MAPPING_MODE_CHOICES:
        raise ValueError(f'Unknown mode {mode}')

    barcode_version = barcode_version.upper()
    if barcode_version not in ['V1', 'V2']:
        raise ValueError(f'Unknown mode {barcode_version}')

    if bismark_ref is not None:
        bismark_ref = pathlib.Path(bismark_ref).absolute()
    else:
        if hisat3n_rna_ref is None or hisat3n_dna_ref is None:
            raise ValueError('bismark_ref is required if hisat3n_rna_ref and hisat3n_dna_ref are not specified.')
        hisat3n_rna_ref = pathlib.Path(hisat3n_rna_ref).absolute()
        hisat3n_dna_ref = pathlib.Path(hisat3n_dna_ref).absolute()

    if mode == 'mct':
        if star_ref is None:
            if hisat3n_rna_ref is None:
                raise ValueError('star_ref or hisat3n_rna_ref is required if mode is mct.')
        else:
            star_ref = pathlib.Path(star_ref).absolute()
        if gtf is None:
            raise ValueError('gtf must be provided when mode is mct.')
        gtf = pathlib.Path(gtf).absolute()

    if chrom_size_path is None:
        raise ValueError('chrom_size_path must be provided.')
    chrom_size_path = pathlib.Path(chrom_size_path).absolute()

    if mode == 'm3c':
        pass

    if mode == '4m':
        if (star_ref is None) and (hisat3n_rna_ref is None):
            raise ValueError('star_ref or hisat3n_rna_ref is required if mode is mct.')
        star_ref = pathlib.Path(star_ref).absolute()

        if gtf is None:
            raise ValueError('gtf must be provided when mode is mct.')
        gtf = pathlib.Path(gtf).absolute()

    genome_fasta = pathlib.Path(genome_fasta).absolute()

    if mode == 'mc':
        if nome:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_nome.ini'
        else:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mc.ini'
        with open(config_path) as f:
            config_content = f.read()
    elif mode == 'mct':
        if nome:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mct-nome.ini'
        else:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mct.ini'
        with open(config_path) as f:
            config_content = f.read()
        if hisat3n_rna_ref is None:
            config_content = config_content.replace('CHANGE_THIS_TO_YOUR_STAR_REFERENCE_DIR', str(star_ref))
        else:
            config_content = config_content.replace('CHANGE_THIS_TO_YOUR_HISAT3N_RNA_REFERENCE',
                                                    str(hisat3n_rna_ref))
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_GENE_ANNOTATION_GTF', str(gtf))
    elif mode == 'm3c':
        config_path = PACKAGE_DIR / 'files/default_config/mapping_config_m3c.ini'
        with open(config_path) as f:
            config_content = f.read()
    elif mode == '4m':
        config_path = PACKAGE_DIR / 'files/default_config/mapping_config_4m.ini'
        with open(config_path) as f:
            config_content = f.read()
        if hisat3n_rna_ref is None:
            config_content = config_content.replace('CHANGE_THIS_TO_YOUR_STAR_REFERENCE_DIR', str(star_ref))
        else:
            config_content = config_content.replace('CHANGE_THIS_TO_YOUR_HISAT3N_RNA_REFERENCE',
                                                    str(hisat3n_rna_ref))
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_GENE_ANNOTATION_GTF', str(gtf))
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_CHROM_SIZE_PATH', str(chrom_size_path))
    else:
        raise

    config_content = config_content.replace('CHANGE_THIS_TO_YOUR_CHROM_SIZE_PATH', str(chrom_size_path))
    config_content = config_content.replace('USE_CORRECT_BARCODE_VERSION_HERE', barcode_version)
    if hisat3n_dna_ref is None:
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_BISMARK_REFERENCE_DIR', str(bismark_ref))
    else:
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_HISAT3N_DNA_REFERENCE', str(hisat3n_dna_ref))
    config_content = config_content.replace('CHANGE_THIS_TO_YOUR_REFERENCE_FASTA', str(genome_fasta))
    print(config_content)
    return



import pandas as pd
from pathlib import Path
import shutil
from .mc_bulk_multigroup_template import MERGE_TEMPLATE, MERGE_EXTRACT_TEMPLATE


'''
the group_file is in csv format with the first column as allcpath and different group the others.
the group_file needs a header.
'''

def merge_bulk_multigroup(group_path, output_path, chrom_size_path, 
                          n_cpu=10, elem_snakegroup_num = 50,
                          cate_snakegroup_num = 10, ):
    

    outdir = Path(output_path)
    outdir.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(group_path, outdir/'GROUP.csv')

    df = pd.read_csv(group_path)
    df = df.rename(columns={df.columns[0]:'_path'})
    sample_cates = df.columns[1:]

    df['_elem'] = pd.factorize(df[sample_cates].astype(str).apply('-'.join, axis=1))[0]
    countdict = df['_elem'].value_counts().to_dict()
    df['_elem'] = df['_elem'].apply(lambda x: f'{x}_{countdict[x]}')

    elem_grp_df = df.groupby('_elem')['_path'].apply(lambda x: x.unique()).to_frame()
    elem_grp_df.index.name = '_sample'
    elem_grp_df['_cate'] = '_elem'
    elem_grp_df = elem_grp_df.reset_index()[['_cate','_sample','_path']]

    df = df[df.columns[1:]].drop_duplicates()

    df['_path'] = output_path+'/_elem/'+df['_elem']+'.allc.tsv.gz'

    cate_grp_df = []
    for cate in sample_cates:
        catedf = df[[cate,'_path']].groupby(cate)['_path'].apply(lambda x: x.unique()).to_frame()
        catedf['_cate'] = cate
        catedf.index.name = '_sample'
        catedf = catedf.reset_index()
        cate_grp_df.append(catedf)
    cate_grp_df = pd.concat(cate_grp_df).reset_index(drop=True)[['_cate','_sample','_path']]   


    def prepare_snakefiles(grp_df, output_path, tag, n_per_snake=None, template=MERGE_TEMPLATE):
        outdir = Path(output_path)
        snkdir = outdir/'snakefiles'
        snkdir.mkdir(exist_ok=True)

        for cate in grp_df['_cate'].unique():
            catedir = outdir/cate
            catedir.mkdir(exist_ok=True)

        for _,(cate,sample,paths) in grp_df.iterrows():
            catedir = outdir/cate
            with open(catedir/f'{sample}.pathlist','w') as f:
                f.write('\n'.join(paths))

        if n_per_snake is None:
            n_per_snake = len(grp_df)

        snk_ids = []
        for i, snkdf in grp_df.groupby(grp_df.index%n_per_snake):
            snk_id = f'{tag}_{i}'

            tocp_df = snkdf[snkdf['_path'].apply(len)==1]
            tomg_df = snkdf[snkdf['_path'].apply(len)>1]

            with open(snkdir/f'{snk_id}.snakefile', 'w') as f:
                f.write(
f'''merge_allc_cpu = {n_cpu}
mcg_context = 'CGN'
chrom_size_path = '{chrom_size_path}'
merge_sample_prefixes = [{','.join("'"+tomg_df['_cate']+'/'+tomg_df['_sample']+"'")}]
copy_sample_prefixes = [{','.join("'"+tocp_df['_cate']+'/'+tocp_df['_sample']+"'")}]
group = "{snk_id}"
'''
                )
                f.write(template)
            snk_ids.append(snk_id)

        return snk_ids

    elem_snk_ids = prepare_snakefiles(elem_grp_df, output_path, 'elem',elem_snakegroup_num, template=MERGE_TEMPLATE)
    cate_snk_ids = prepare_snakefiles(cate_grp_df, output_path, 'cate',cate_snakegroup_num, template=MERGE_EXTRACT_TEMPLATE)

    def prepare_commands(snake_ids):
        cmds = [f'snakemake  -d {outdir.resolve()} --snakefile {outdir.resolve()}/snakefiles/{snkid}.snakefile '
                f'-j {n_cpu} --default-resources mem_mb=100 --resources mem_mb=1000 --rerun-incomplete' \
                for snkid in snake_ids]
        return cmds

    
    
    with open(outdir/'run_snakemake_cmds_1.txt', 'w') as f:
        f.write('\n'.join(prepare_commands(elem_snk_ids)))
    with open(outdir/'run_snakemake_cmds_2.txt', 'w') as f:
        f.write('\n'.join(prepare_commands(cate_snk_ids)))    



from collections import defaultdict
import re
import pysam
import pandas as pd


def read_mc_level(read, frac=True, nome=False):
    bismark_tag = read.get_tag('XM')
    if nome:
        m_c = 0
        normal_c = 0
        seq = read.seq.upper()
        read_length = len(seq)
        for pos, xm_base in enumerate(bismark_tag):
            if xm_base in '.ZzUu':
                # skip unrelated base (.), CpG (Zz), CpUnknown (Uu)
                continue
            # Skip GpC
            try:
                if read.is_reverse:
                    if (pos == read_length) or (read.seq[pos + 1] == 'C'):
                        continue
                else:
                    if (pos == 0) or (read.seq[pos - 1] == 'G'):
                        continue
            except IndexError:
                # start or end of the read
                continue
            if xm_base in 'xh':
                normal_c += 1
            elif xm_base in 'XH':
                m_c += 1
            else:
                pass
    else:
        m_c = bismark_tag.count('X') + bismark_tag.count('H')
        normal_c = bismark_tag.count('x') + bismark_tag.count('h')

    total_c = m_c + normal_c
    if total_c == 0:
        return 0, 0
    else:
        if frac:
            read_mc_rate = m_c / total_c
            return read_mc_rate, total_c
        else:
            return m_c, total_c


def select_dna_reads_normal(input_bam,
                            output_bam,
                            mc_rate_max_threshold=0.5,
                            cov_min_threshold=3,
                            nome=False):
    read_profile_dict = defaultdict(int)
    # init dict to make sure the series has something
    read_profile_dict[(50, 50)] = 0
    with pysam.AlignmentFile(input_bam) as f:
        with pysam.AlignmentFile(output_bam, header=f.header,
                                 mode='wb') as out_f:
            for read in f:
                mc_frac, cov = read_mc_level(read, nome=nome)
                read_profile_dict[(int(100 * mc_frac), cov)] += 1

                # split reads
                if (mc_frac > mc_rate_max_threshold) or (cov <
                                                         cov_min_threshold):
                    continue
                out_f.write(read)
    with open(str(output_bam) + '.reads_profile.csv', 'w') as stat_f:
        stat_f.write('mc_frac,cov,count\n')
        for (mc_frac, cov), count in read_profile_dict.items():
            stat_f.write(f'{mc_frac},{cov},{count}\n')
    return


def select_dna_reads_split_reads(input_bam,
                                 output_bam,
                                 mc_rate_max_threshold=0.5,
                                 cov_min_threshold=3,
                                 nome=False):
    splited_read_name_pattern = re.compile('.+-[lrm]$')

    # first pass: determine read methylation level
    read_level_mcs = defaultdict(int)
    read_level_covs = defaultdict(int)
    with pysam.AlignmentFile(input_bam) as f:
        for read in f:
            mc, cov = read_mc_level(read, frac=False, nome=nome)
            read_name = read.qname
            if splited_read_name_pattern.search(read_name):
                read_level_mcs[read_name[:-2]] += mc
                read_level_covs[read_name[:-2]] += cov
            else:
                read_level_mcs[read_name] += mc
                read_level_covs[read_name] += cov
    read_level_data = pd.DataFrame({
        'mc': read_level_mcs,
        'cov': read_level_covs
    })
    read_level_data['mc_frac'] = read_level_data['mc'] / (read_level_data['cov'] +
                                                       0.001)
    read_level_data['mc_frac'] = (read_level_data['mc_frac'] * 100).astype(int)
    if read_level_data.shape[0] == 0:
        # in case there is no read at all:
        with open(f'{output_bam}.reads_profile.csv', 'w') as f:
            f.write('mc_frac,cov,count\n')
            f.write('0,1,0\n')
    else:
        profile = read_level_data.groupby('mc_frac')['cov'].value_counts()
        profile.name = 'count'
        profile = profile.reset_index()
        profile.to_csv(f'{output_bam}.reads_profile.csv', index=None)

    # filter reads
    use_reads = read_level_data[
        (read_level_data['mc_frac'] < mc_rate_max_threshold)
        & (read_level_data['cov'] >= cov_min_threshold)].index.tolist()
    use_reads = set(use_reads)
    del read_level_data

    # second pass: write passed reads
    with pysam.AlignmentFile(input_bam) as f:
        with pysam.AlignmentFile(output_bam, header=f.header,
                                 mode='wb') as out_f:
            for read in f:
                read_name = read.qname
                if (read_name in use_reads) or (read_name[:-2] in use_reads):
                    # read name or read name without suffix
                    out_f.write(read)
    return


def select_dna_reads(input_bam,
                     output_bam,
                     mc_rate_max_threshold=0.5,
                     cov_min_threshold=3,
                     nome=False,
                     assay_type='mc'):
    if assay_type == 'mc':
        select_dna_reads_normal(input_bam,
                                output_bam,
                                mc_rate_max_threshold=mc_rate_max_threshold,
                                cov_min_threshold=cov_min_threshold,
                                nome=nome)
    elif assay_type == 'm3c':
        select_dna_reads_split_reads(input_bam,
                                     output_bam,
                                     mc_rate_max_threshold=mc_rate_max_threshold,
                                     cov_min_threshold=cov_min_threshold,
                                     nome=nome)
    else:
        raise ValueError(f'Unknown assay_type {assay_type}.')
    return



"""
Demultiplex pipeline
"""

import logging
import pathlib
import re
import subprocess

import pandas as pd

import cemba_data
from .fastq_dataframe import make_fastq_dataframe
from ..mapping.pipelines import make_snakefile, prepare_run, validate_mapping_config
from ..utilities import snakemake, get_configuration
from ..hisat3n import make_snakefile_hisat3n

# logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])


def _demultiplex(fastq_pattern, output_dir, barcode_version, cpu):
    """
    Input raw FASTQ file pattern
    1. automatically parse the name to generate fastq dataframe
    2. use the FASTQ dataframe to generate dir structure for each uid
    3. generate snakefile for demultiplexing each uid
    4. generate final snakefile in output_dir
    5. execute snakefile

    Parameters
    ----------
    fastq_pattern
    output_dir
    barcode_version
    cpu

    Returns
    -------

    """
    output_dir = pathlib.Path(output_dir).absolute()

    # make fastq dataframe
    fastq_df = make_fastq_dataframe(fastq_pattern,
                                    barcode_version=barcode_version,
                                    output_path=output_dir / 'stats' /
                                                'fastq_dataframe.csv')

    # judge the type of V2 barcoding
    if (barcode_version == 'V2') and (fastq_df['multiplex_group'].unique().size == 1):
        # in this case, the six multiplex group is indexed by a single PCR index
        print(
            'Detect only single multiplex group in each plate, will use V2-single mode.'
        )
        barcode_version = 'V2-single'
        # the resulting demultiplexed dir will have all the 384 cells.
        # I added a patch in the end to split them into multiplex groups that each sub-dir only contain 64 cells.

    # prepare UID sub dir
    snakefile_list = []
    total_stats_list = []
    rule_count = 0
    for uid, uid_df in fastq_df.groupby('uid'):
        # determine index file path
        if barcode_version == 'V1':
            random_index_fasta_path = str(PACKAGE_DIR /
                                          'files/random_index_v1.fa')
        elif barcode_version == 'V2':
            multiplex_group = uid.split('-')[-2]
            random_index_fasta_path = str(
                PACKAGE_DIR / 'files/random_index_v2/'
                              f'random_index_v2.multiplex_group_{multiplex_group}.fa')
        elif barcode_version == 'V2-single':
            random_index_fasta_path = str(PACKAGE_DIR /
                                          'files/random_index_v2/'
                                          f'random_index_v2.fa')
        else:
            raise ValueError(f'Got unknown barcode version {barcode_version}.')

        # create a directory for each uid, within this UID, do multiplex and lane merge
        uid_output_dir = output_dir / uid
        uid_output_dir.mkdir(exist_ok=True)
        lane_files_dir = uid_output_dir / 'lanes'
        lane_files_dir.mkdir(exist_ok=True)

        # standardize input fastq name for easier parsing
        raw_dir = uid_output_dir / 'raw'
        raw_dir.mkdir(exist_ok=True)
        for _, row in uid_df.iterrows():
            uid, read_type, lane, old_path = row[[
                'uid', 'read_type', 'lane', 'fastq_path'
            ]]
            new_path = raw_dir / f'{uid}+{lane}+{read_type}.fq.gz'
            subprocess.run(['ln', '-s', old_path, new_path], check=True)
        lanes = list(uid_df['lane'].unique())
        name_str = '{{name}}'

        # make snakefile
        stats_out_list = [
            f'{lane_files_dir}/{uid}-{lane}.demultiplex.stats.txt'
            for lane in lanes
        ]
        total_stats_list += stats_out_list
        rules = ""
        for lane in lanes:
            snake_file_template = f"""
rule demultiplex_{rule_count}:
    input:
        r1_in = f'{raw_dir}/{uid}+{lane}+R1.fq.gz',
        r2_in = f'{raw_dir}/{uid}+{lane}+R2.fq.gz'
    params:
        # Note that you have to use a function to deactivate automatic wildcard expansion 
        # in params strings, e.g., `lambda wildcards: ...`.
        # here the r1/2_out have to have the name_str
        r1_out = lambda wildcards: f'{lane_files_dir}/{uid}-{lane}-{name_str}-R1.fq.gz',
        r2_out = lambda wildcards: f'{lane_files_dir}/{uid}-{lane}-{name_str}-R2.fq.gz'
    output:
        stats_out = '{lane_files_dir}/{uid}-{lane}.demultiplex.stats.txt'
    shell:
        "cutadapt -Z -e 0.01 --no-indels -g file:{random_index_fasta_path} "
        "-o {{params.r1_out}} -p {{params.r2_out}} {{input.r1_in}} {{input.r2_in}} > {{output.stats_out}}"
    """
            rule_count += 1
            rules += snake_file_template

        snake_file_path = lane_files_dir / 'Snakefile'
        with open(snake_file_path, 'w') as f:
            f.write(rules)
        snakefile_list.append(f'{uid}/lanes/Snakefile')

    # make final snakefile for demultiplex step
    final_rules = ''
    for path in snakefile_list:
        final_rules += f'include: "{path}"\n'
    # final rules
    final_rules += f"""
rule final:
    input: {total_stats_list}
"""
    final_snake_path = output_dir / 'Snakefile_demultiplex'
    with open(final_snake_path, 'w') as f:
        f.write(final_rules)

    print('Demultiplexing raw FASTQ')
    snakemake(workdir=output_dir, snakefile=final_snake_path, cores=cpu)
    return barcode_version


def _merge_lane(output_dir, cpu):
    output_dir = pathlib.Path(output_dir).absolute()
    fastq_df = pd.read_csv(output_dir / 'stats' / 'fastq_dataframe.csv')
    snakefile_list = []
    total_output_list = []
    rule_uid = 0
    # prepare snakefile in each uid
    for uid in fastq_df['uid'].unique():
        uid_output_dir = output_dir / uid
        lanes_dir = uid_output_dir / 'lanes'
        fastq_dir = uid_output_dir / 'fastq'
        fastq_dir.mkdir(exist_ok=True)

        # prepare demultiplex results cell_fastq_df
        records = []
        for path in lanes_dir.glob('*fq.gz'):
            *uid, lane, index_name, read_type = path.name[:-6].split('-')
            uid = '-'.join(uid)
            cell_id = f'{uid}-{index_name}'
            records.append([cell_id, lane, read_type, str(path)])
        cell_fastq_df = pd.DataFrame(
            records,
            columns=['cell_id', 'index_name', 'read_type', 'fastq_path'])

        # prepare snakefile for each cell_id * read_type
        rules = ''
        output_paths = []
        for (cell_id, read_type), sub_df in cell_fastq_df.groupby(
                ['cell_id', 'read_type']):
            input_paths = list(sub_df['fastq_path'])
            output_path = fastq_dir / f'{cell_id}-{read_type}.fq.gz'

            snake_file_template = f"""
rule merge_{rule_uid}:
    input: 
        {input_paths}
    output: 
        "{output_path}"
    shell:
        "gzip -cd {{input}} | gzip -5 > {{output}} && rm -f {{input}}"

"""
            rule_uid += 1
            rules += snake_file_template
            output_paths.append(str(output_path))

        snakefile_path = uid_output_dir / 'Snakefile'
        with open(snakefile_path, 'w') as f:
            f.write(rules)
        snakefile_list.append(snakefile_path)
        total_output_list += output_paths

    # prepare final snakefile
    final_rules = ''
    for path in snakefile_list:
        final_rules += f'include: "{path}"\n'
    # final rules
    final_rules += f"""
rule final:
    input: {total_output_list}
"""
    final_snake_path = output_dir / 'Snakefile_merge_lane'
    with open(final_snake_path, 'w') as f:
        f.write(final_rules)

    print('Merging lanes to get cell FASTQ')
    snakemake(workdir=output_dir, snakefile=final_snake_path, cores=cpu)
    return


def _parse_index_fasta(fasta_path):
    records = {}
    with open(fasta_path) as f:
        key_line = True
        for line in f:
            if key_line:
                key = line.lstrip('>').rstrip('\n')
                key_line = False
            else:
                value = line.lstrip('^').rstrip('\n')
                records[key] = value
                key_line = True
    return records


def _read_cutadapt_result(stat_path):
    """
    Parser of cutadapt output
    """
    with open(stat_path) as f:
        p = re.compile(
            r"Sequence: .+; Type: .+; Length: \d+; Trimmed: \d+ times")
        series = []
        total_pairs = -1
        for line in f:
            if line.startswith('Total read pairs processed'):
                total_pairs = line.split(' ')[-1]
                total_pairs = int(''.join(re.compile(r'\d').findall(total_pairs)))

            m = p.search(line)
            if m is not None:
                result_dict = {}
                for i in m.group().split('; '):
                    k, v = i.split(': ')
                    result_dict[k] = v
                result_series = pd.Series(result_dict)
                series.append(result_series)
        total_df = pd.DataFrame(series)
        total_df['Trimmed'] = total_df['Trimmed'].apply(
            lambda c: c.split(' ')[0]).astype(int)
        total_df['TotalPair'] = total_pairs
        total_df['Ratio'] = total_df['Trimmed'] / total_pairs
    return total_df


def _summarize_demultiplex(output_dir, barcode_version):
    output_dir = pathlib.Path(output_dir).absolute()
    output_path = output_dir / 'stats' / 'demultiplex.stats.csv'
    barcode_version = barcode_version.upper()

    # get index info
    if barcode_version == 'V1':
        random_index_fasta_path = str(PACKAGE_DIR / 'files/random_index_v1.fa')
    elif barcode_version == 'V2':
        # here we don't need to worry about the multiplex_group issue,
        # because we just need a index_name to index_seq map
        # we've considered this during demultiplex
        random_index_fasta_path = str(
            PACKAGE_DIR / 'files/random_index_v2/random_index_v2.fa')
    else:
        raise ValueError(
            f'Unknown version name {barcode_version} in multiplexIndex section of the config file.'
        )
    index_seq_dict = _parse_index_fasta(random_index_fasta_path)
    index_name_dict = {v: k for k, v in index_seq_dict.items()}

    # read the demultiplex stats, its per lane, so need to sum up lane together of each uid and index name
    # but R1 R2 is demultiplexed together, so this table don't separate R1 R2
    stat_list = []
    stat_path_list = list(output_dir.glob('*/lanes/*demultiplex.stats.txt'))
    for path in stat_path_list:
        single_df = _read_cutadapt_result(path)
        *uid, suffix = path.name.split('-')
        lane = suffix.split('.')[0]
        uid = '-'.join(uid)
        single_df['uid'] = uid
        single_df['lane'] = lane
        single_df['index_name'] = single_df['Sequence'].map(index_name_dict)
        assert single_df['index_name'].isna().sum() == 0
        stat_list.append(single_df)
    total_demultiplex_stats = pd.concat(stat_list)

    # calculate cell level table
    total_demultiplex_stats['cell_id'] = total_demultiplex_stats[
                                             'uid'] + '-' + total_demultiplex_stats['index_name']

    cell_table = total_demultiplex_stats.groupby('cell_id').agg({
        'Trimmed':
            'sum',
        'TotalPair':
            'sum',
        'index_name':
            lambda i: i.unique()[0],
        'uid':
            lambda i: i.unique()[0]
    })
    cell_table.rename(columns={
        'Trimmed': 'CellInputReadPairs',
        'TotalPair': 'MultiplexedTotalReadPairs',
        'index_name': 'IndexName',
        'uid': 'UID'
    },
        inplace=True)
    cell_table['CellBarcodeRate'] = cell_table[
                                        'CellInputReadPairs'] / cell_table['MultiplexedTotalReadPairs']
    cell_table['BarcodeVersion'] = barcode_version
    cell_table.to_csv(output_path)
    return


def _final_cleaning(output_dir):
    """
    remove intermediate files from demultiplex
    """
    output_dir = pathlib.Path(output_dir)

    delete_patterns = [
        f'Snakefile_*', '*/lanes', '*/raw', '*/Snakefile',
        '*/fastq/*-unknown-R*.fq.gz', '.snakemake'
    ]

    total_paths = []
    for pattern in delete_patterns:
        total_paths += list(map(str, output_dir.glob(pattern)))

    subprocess.run(['rm', '-rf'] + total_paths, check=True)
    return


def _skip_abnormal_fastq_pairs(output_dir):
    demultiplex_df = pd.read_csv(output_dir / 'stats/demultiplex.stats.csv',
                                 index_col=0)
    config = get_configuration(output_dir / 'mapping_config.ini')
    total_read_pairs_min = int(config['total_read_pairs_min'])
    total_read_pairs_max = int(config['total_read_pairs_max'])

    too_large = demultiplex_df['CellInputReadPairs'] > total_read_pairs_max
    too_small = demultiplex_df['CellInputReadPairs'] < total_read_pairs_min
    judge = too_small | too_large
    unmapped_cells = demultiplex_df[judge]
    print(
        f'Skip {too_small.sum()} cells due to too less input read pairs (< {total_read_pairs_min})'
    )
    print(
        f'Skip {too_large.sum()} cells due to too large input read pairs (> {total_read_pairs_max})'
    )

    for cell_id, row in unmapped_cells.iterrows():
        uid = row['UID']
        skipped_dir = output_dir / uid / 'fastq/skipped/'
        skipped_dir.mkdir(exist_ok=True, parents=True)

        # move both R1 R2 to skipped files, it will not be included in Snakefile
        for read_type in ['R1', 'R2']:
            fastq_path = output_dir / uid / f'fastq/{cell_id}-{read_type}.fq.gz'
            new_path = skipped_dir / f'{cell_id}-{read_type}.fq.gz'
            # if CellInputReadPairs = 0, the FASTQ file do not actually exist, but it does have a row in metadata.
            if fastq_path.exists():
                subprocess.run(['mv', str(fastq_path),
                                str(new_path)],
                               check=True)

    # save UID total input reads, for command order
    uid_order = demultiplex_df[~judge].groupby(
        'UID')['CellInputReadPairs'].sum().sort_values(ascending=False)
    uid_order.to_csv(output_dir / 'stats/UIDTotalCellInputReadPairs.csv',
                     header=False)
    return


def _reformat_v2_single(output_dir):
    # fix names and dir structure
    output_dir = pathlib.Path(output_dir)
    fastq_paths = list(output_dir.glob('*/fastq/*gz'))
    for path in fastq_paths:
        plate, cur_group, pcr_index, random_index, suffix = path.name.split(
            '-')
        col = int(random_index[1:])
        real_group = ((col - 1) % 12) // 2 + 1
        new_path = str(path).replace(f'{plate}-{cur_group}-{pcr_index}',
                                     f'{plate}-{real_group}-{pcr_index}')
        path = str(path)
        if path == new_path:
            continue
        pathlib.Path(new_path).parent.mkdir(exist_ok=True, parents=True)
        subprocess.run(['mv', path, new_path])

    # fix names in the demultiplex stats
    cell_stats = pd.read_csv(output_dir / 'stats/demultiplex.stats.csv',
                             index_col=0)
    new_stats = []
    for cell_id, row in cell_stats.iterrows():
        real_group = ((int(row['IndexName'][1:]) - 1) % 12) // 2 + 1
        plate, _, pcr_index, random_index = cell_id.split('-')
        row.name = f'{plate}-{real_group}-{pcr_index}-{random_index}'
        row['UID'] = f'{plate}-{real_group}-{pcr_index}'
        new_stats.append(row)
    new_stats = pd.DataFrame(new_stats)
    new_stats.to_csv(output_dir / 'stats/demultiplex.stats.csv')
    return


SUPPORTED_TECHNOLOGY = ['mc', 'mct', 'm3c']


def demultiplex_pipeline(fastq_pattern, output_dir, config_path, cpu, aligner):
    cpu = int(cpu)
    merge_cpu = min(48, cpu)
    demultiplex_cpu = min(32, cpu)

    output_dir = pathlib.Path(output_dir).absolute()
    if output_dir.exists():
        raise FileExistsError(
            'output_dir already exists, to prevent conflicts, '
            'use another output_dir or delete the existing output_dir first.')
    else:
        output_dir.mkdir(parents=True)
        (output_dir / 'stats').mkdir()

    config = get_configuration(config_path)
    suffix = pathlib.Path(config_path).name.split('.')[-1]
    new_config_path = output_dir / f'mapping_config.{suffix}'
    subprocess.run(f'cp {config_path} {new_config_path}',
                   shell=True,
                   check=True)
    barcode_version = config['barcode_version']

    # validate config file first before demultiplex
    validate_mapping_config(output_dir)

    demultiplex_version = _demultiplex(fastq_pattern=fastq_pattern,
                                       output_dir=output_dir,
                                       barcode_version=barcode_version,
                                       cpu=demultiplex_cpu)
    _merge_lane(output_dir=output_dir, cpu=merge_cpu)
    _summarize_demultiplex(output_dir=output_dir,
                           barcode_version=barcode_version)
    _final_cleaning(output_dir=output_dir)
    if demultiplex_version == 'V2-single':
        print('Reformat directories to separate multiplex groups')
        _reformat_v2_single(output_dir=output_dir)
    _skip_abnormal_fastq_pairs(output_dir=output_dir)

    if aligner.lower() == 'bismark':
        make_snakefile(output_dir=output_dir)

        # this is just a convenient step, so I fix the parameters here
        # users should change the resulting batch submission
        # or generate by themselves if they want different setting.
        prepare_run(output_dir)
    elif aligner.lower() in ('hisat3n', 'hisat-3n', 'hisat_3n', 'hisat'):
        make_snakefile_hisat3n(output_dir=output_dir)
    else:
        ValueError(f'Unsupported aligner {aligner}, aligner must be either "hisat3n" or "bismark"')
    return


def update_snakemake(output_dir):
    """When mapping_config.ini is updated, or the output_dir path changed,
    use this function to update snakefile and snakemake commands."""
    output_dir = pathlib.Path(output_dir).absolute()

    if pathlib.Path(output_dir / 'snakemake/hisat3n').exists():
        # hisat3n pipeline
        make_snakefile_hisat3n(output_dir=output_dir)
    else:
        make_snakefile(output_dir=output_dir)
        prepare_run(output_dir)
    return



import pandas as pd
import h5py
import numpy as np
import pyBigWig
import subprocess
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed
from ALLCools.utilities import parse_chrom_size


def dump_single_snap(output_prefix, snap_path, barcode_to_cluster):
    cluster_handle_dict = {cluster: open(f'{output_prefix}_{cluster}.bed', 'w')
                           for cluster in barcode_to_cluster.unique()}
    cluster_n_frags = defaultdict(int)
    with h5py.File(snap_path, 'r') as snap:
        barcode = pd.Series(np.array(snap['/BD/name']).astype(str))
        barcode_length = np.array(snap['/FM/barcodeLen'])
        barcode_pos = np.array(snap['/FM/barcodePos']) - 1
        barcode_table = pd.DataFrame({
            'barcode': barcode,
            'length': barcode_length,
            'start': barcode_pos
        })
        barcode_table['end'] = barcode_table['start'] + barcode_table['length']
        barcode_table['cluster'] = barcode.map(barcode_to_cluster)
        barcode_table.dropna(inplace=True)

        for i, (_, row) in enumerate(barcode_table.iterrows()):
            *_, start, end, cluster = row
            frag_chrom = pd.Series(
                np.array(snap['/FM/fragChrom'][start:end]).astype(str))
            frag_start = pd.Series(np.array(snap['/FM/fragStart'][start:end]))
            frag_length = pd.Series(np.array(snap['/FM/fragLen'][start:end]))
            frag_table = pd.DataFrame({
                'chrom': frag_chrom,
                'start': frag_start,
                'end': frag_start + frag_length
            })
            cluster_handle_dict[cluster].write(frag_table.to_csv(sep='\t',
                                                                 header=None,
                                                                 index=None))
            cluster_n_frags[cluster] += frag_table.shape[0]

    for h in cluster_handle_dict.values():
        h.close()
    return cluster_n_frags


def fragments_to_bigwig(output_prefix, cluster, chrom_size_path, bw_bin_size=10, scale=None):
    # concat bed
    subprocess.run(
        f'cat {output_prefix}_*_{cluster}.bed > {output_prefix}_{cluster}.bed && '
        f'rm -f {output_prefix}_*_{cluster}.bed',
        shell=True,
        check=True)
    # sort bed
    subprocess.run(f'sort -k 1,1 -k2,2n {output_prefix}_{cluster}.bed '
                   f'> {output_prefix}_{cluster}.sorted.bed && '
                   f'rm -f {output_prefix}_{cluster}.bed',
                   shell=True,
                   check=True)
    # bed to bedgraph
    subprocess.run(f'bedtools genomecov -i {output_prefix}_{cluster}.sorted.bed '
                   f'-g {chrom_size_path} -bg > {output_prefix}_{cluster}.bedgraph',
                   shell=True,
                   check=True)
    bg_path = f'{output_prefix}_{cluster}.bedgraph'
    bw_path = f'{output_prefix}_{cluster}.bw'

    bg_iter = pd.read_csv(bg_path,
                          sep='\t',
                          header=None,
                          names=['chrom', 'start', 'end', 'count'],
                          chunksize=100000)
    total_wigs = []
    for bg in bg_iter:
        bg['start'] = bg['start'] // bw_bin_size
        wig_values = bg.groupby(['chrom',
                                 'start'])['count'].mean().reset_index()
        total_wigs.append(wig_values)
    total_wigs = pd.concat(total_wigs)
    total_wigs = total_wigs.groupby(['chrom',
                                     'start'])['count'].mean().reset_index()

    chrom_sizes = parse_chrom_size(chrom_size_path)
    chrom_sizes_list = [(k, v) for k, v in chrom_sizes.items()]

    with pyBigWig.open(bw_path, 'w') as bw_out:
        bw_out.addHeader(chrom_sizes_list)
        for chrom in chrom_sizes.keys():
            chrom_df = total_wigs[total_wigs['chrom'] == chrom].sort_values(
                'start')
            if chrom_df.shape[0] == 0:
                continue
            if scale is None:
                values = chrom_df['count'].astype(float)
            else:
                values = chrom_df['count'].astype(float) / scale
            bw_out.addEntries(chrom, (chrom_df['start'] * bw_bin_size).tolist(),
                              values=values.tolist(),
                              span=bw_bin_size)

    # remove bed graph
    subprocess.run(f'rm -f {bg_path}', shell=True, check=True)
    # gzip bed
    subprocess.run(f'gzip {output_prefix}_{cluster}.sorted.bed', shell=True, check=True)
    return


def prepare_atac_bulk(cluster_table,
                      snap_table,
                      output_prefix,
                      chrom_size_path,
                      bw_bin_size=10,
                      cpu=1):
    """
    Merge snATAC single cell fragments stored in SNAP files into pseudo bulk and generate bigwig.

    Parameters
    ----------
    cluster_table
        Three-column CSV,
        1) sample id (SNAP sample, not cell)
        2) cell barcode within a sample
        3) cluster label
    snap_table
        Two-column CSV,
        1) sample id, matching cluster table col 1;
        2) path to SNAP file
    output_prefix
        overall output prefix
    chrom_size_path
        path to chrom sizes file
    bw_bin_size
        bigwig bin size
    cpu
        parallel CPU

    Returns
    -------

    """
    cluster_table = pd.read_csv(cluster_table, index_col=None, header=None,
                                names=['sample', 'barcode', 'cluster'])
    snap_table = pd.read_csv(snap_table, index_col=0, header=None, squeeze=True)

    sample_clusters = {s: c.set_index('barcode')['cluster']
                       for s, c in cluster_table.groupby('sample')}

    # dump fragments
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for sample, clusters in sample_clusters.items():
            _output_prefix = f'{output_prefix}_{sample}'
            f = exe.submit(dump_single_snap,
                           output_prefix=_output_prefix,
                           snap_path=snap_table[sample],
                           barcode_to_cluster=clusters)
            futures[f] = sample

        total_n_frags = defaultdict(int)
        for f in as_completed(futures):
            sample = futures[f]
            print(f'{sample} returned')
            cluster_n_frags = f.result()
            for k, v in cluster_n_frags.items():
                total_n_frags[k] += v

    million_reads_scaler = {k: v / 1000000 for k, v in total_n_frags.items()}
    total_clusters = pd.concat(sample_clusters.values()).unique()
    # generate bed and bigwig
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for cluster in total_clusters:
            f = exe.submit(fragments_to_bigwig,
                           output_prefix,
                           cluster,
                           chrom_size_path=chrom_size_path,
                           bw_bin_size=bw_bin_size,
                           scale=million_reads_scaler[cluster])
            futures[f] = cluster

        for f in as_completed(futures):
            cluster = futures[f]
            print(f'{cluster} returned')
            f.result()



import pandas as pd
import pathlib
import yaml
from cooler.util import binnify, read_chromsizes
import subprocess
from statsmodels.stats.multitest import multipletests

import cemba_data

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])
DMRSEQ_TEMPLATE = PACKAGE_DIR / 'dmr/dmrseq/DMRseq.ipynb'


def prepare_snakemake(allc_table_path,
                      output_dir,
                      chrom_sizes_path,
                      template_path,
                      chroms=None,
                      test_covariate='group',
                      match_covariate=None,
                      adjust_covariate=None,
                      cutoff=0.1,
                      min_num_region=3,
                      smooth=True,
                      bp_span=1000,
                      min_in_span=30,
                      max_gap_smooth=2500,
                      max_gap=1000,
                      verbose=True,
                      max_perms=10,
                      stat="stat",
                      block=False,
                      block_size=5000,
                      chrs_per_chunk=1,
                      cpu=40,
                      chunk_size=5000000000):
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    chrom_sizes = read_chromsizes(chrom_sizes_path).reindex(chroms)
    if chroms is None:
        chroms = chrom_sizes.index.tolist()
    bins = binnify(chrom_sizes.loc[chroms], binsize=chunk_size)
    regions = []
    for _, (chrom, start, end) in bins.iterrows():
        region = f'{chrom}:{start}-{end}'
        regions.append(region)

    for region in regions:
        config_path = f'{output_dir}/{region}.yaml'
        parameters = {
            'region': region,
            'allc_table_path': allc_table_path,
            'test_covariate': test_covariate,
            'match_covariate': match_covariate,
            'adjust_covariate': adjust_covariate,
            'cutoff': cutoff,
            'min_num_region': min_num_region,
            'smooth': smooth,
            'bp_span': bp_span,
            'min_in_span': min_in_span,
            'max_gap_smooth': max_gap_smooth,
            'max_gap': max_gap,
            'verbose': verbose,
            'max_perms': max_perms,
            'stat': stat,
            'block': block,
            'block_size': block_size,
            'chrs_per_chunk': chrs_per_chunk,
            'cpu': cpu
        }
        with open(config_path, 'w') as f:
            f.write(yaml.dump(parameters))

    snakefile = f"""
regions = {regions}
rule main:
    input:
        expand('{{region}}.DMR.hdf', region=regions)

rule papermill:
    input:
        nb='{template_path}',
        config='{{region}}.yaml'
    output:
        nb='{{region}}.ipynb',
        data='{{region}}.DMR.hdf'
    threads:
        1 #{cpu}
    shell:
        'papermill {{input.nb}} {{output.nb}} -f {{input.config}} && sleep 10'
"""
    snakefile_path = f'{output_dir}/Snakefile'
    with open(snakefile_path, 'w') as f:
        f.write(snakefile)
    return regions


def run_dmrseq(allc_table_path,
               output_dir,
               study_name,
               chrom_sizes_path,
               chroms=None,
               test_covariate='group',
               match_covariate=None,
               adjust_covariate=None,
               cutoff=0.1,
               min_num_region=3,
               smooth=True,
               bp_span=1000,
               min_in_span=30,
               max_gap_smooth=2500,
               max_gap=1000,
               verbose=True,
               max_perms=10,
               stat="stat",
               block=False,
               block_size=5000,
               chrs_per_chunk=1,
               cpu=40,
               template_path='default',
               chunk_size=50000000):
    # prepare template
    if template_path == 'default':
        template_path = DMRSEQ_TEMPLATE
    allc_table_path = str(pathlib.Path(allc_table_path).absolute())

    # prepare snakemake
    output_dir = pathlib.Path(output_dir).absolute()
    this_study_dir = output_dir / f'{study_name}_DMRseq'
    regions = prepare_snakemake(allc_table_path=allc_table_path,
                                output_dir=this_study_dir,
                                chrom_sizes_path=chrom_sizes_path,
                                template_path=template_path,
                                chroms=chroms,
                                test_covariate=test_covariate,
                                match_covariate=match_covariate,
                                adjust_covariate=adjust_covariate,
                                cutoff=cutoff,
                                min_num_region=min_num_region,
                                smooth=smooth,
                                bp_span=bp_span,
                                min_in_span=min_in_span,
                                max_gap_smooth=max_gap_smooth,
                                max_gap=max_gap,
                                verbose=verbose,
                                max_perms=max_perms,
                                stat=stat,
                                block=block,
                                block_size=block_size,
                                chrs_per_chunk=chrs_per_chunk,
                                cpu=cpu,
                                chunk_size=chunk_size)
    # the ipykernel package raise "zmq.error.ZMQError: Address already in use" due to parallel execution,
    # restart likely solve the problem.
    snakemake_cmd = f'snakemake -d {this_study_dir} --snakefile {this_study_dir}/Snakefile ' \
                    f'-j {cpu} --default-resources mem_mb=100 --resources mem_mb={int(5000 * cpu)} ' \
                    f'--restart-times 3'
    subprocess.run(snakemake_cmd, shell=True, check=True)

    total_dmrs = []
    for region in regions:
        try:
            total_dmrs.append(pd.read_hdf(f'{this_study_dir}/{region}.DMR.hdf'))
        except ValueError:
            pass
    # happens when there is no sites to test in dmrseq, and output file is empty
    total_dmrs = pd.concat(total_dmrs)
    # recalculate FDR
    # some pval in the dmrseq result are nan
    total_dmrs = total_dmrs.loc[~total_dmrs['pval'].isna()].copy()
    _, fdr, *_ = multipletests(total_dmrs['pval'], method='fdr_bh')
    total_dmrs['qval'] = fdr
    total_dmrs.to_hdf(f'{this_study_dir}/{study_name}_DMR.hdf', key='data', format="table")

    for region in regions:
        subprocess.run(f'rm -f {this_study_dir}/{region}.DMR.hdf',
                       shell=True)
        subprocess.run(f'rm -f {this_study_dir}/*.yaml', shell=True)
        subprocess.run(
            f'rm -rf {this_study_dir}/.snakemake {this_study_dir}/Snakefile',
            shell=True)
    return



import pysam
import pandas as pd
import glob
import subprocess
from concurrent.futures import ProcessPoolExecutor, as_completed
import os


def merge_single_bam(bam_path, cell_id_to_cluster, output_prefix, header_dict):
    header = pysam.AlignmentHeader.from_dict(header_dict)
    clusters = set(cell_id_to_cluster.values())
    cluster_read_counts = {c: 0 for c in clusters}

    # write reads by cluster
    with pysam.AlignmentFile(bam_path, "rb") as bam_file:
        # open BAM handles for each cluster
        cluster_handles = {}
        for cluster in clusters:
            cluster_handles[cluster] = pysam.AlignmentFile(
                f'{output_prefix}_{cluster}.bam', "wb", header=header)

        for read in bam_file:
            cell_id = read.get_tag('RG')
            try:
                cluster = cell_id_to_cluster[cell_id]
                # this removes RG tag
                read.set_tag('RG', None)
                cluster_handles[cluster].write(read)
                cluster_read_counts[cluster] += 1
            except KeyError:
                continue

    # close handles
    for handle in cluster_handles.values():
        handle.close()

    # delete empty out_bam
    for cluster, count in cluster_read_counts.items():
        bam_path = f'{output_prefix}_{cluster}.bam'
        if count == 0:
            subprocess.run(['rm', '-rf', bam_path])
    return cluster_read_counts


def merge_mct_cluster_bam(cell_id_to_cluster_path,
                          bam_list_path,
                          output_prefix,
                          cpu=10):
    cell_id_to_cluster = pd.read_csv(
        cell_id_to_cluster_path,
        index_col=0,
        header=None,
        squeeze=True).to_dict()
    bam_paths = pd.read_csv(bam_list_path, header=None, squeeze=True).tolist()

    # get header
    with pysam.AlignmentFile(bam_paths[0]) as bam:
        header_dict = bam.header.as_dict()
        # remove cell specific info
        keys_to_delete = ['PG', 'RG', 'CO']
        for k in keys_to_delete:
            if k in header_dict:
                del header_dict[k]

    clusters = set(cell_id_to_cluster.values())
    total_cluster_read_counts = {c: 0 for c in clusters}

    # merge single bam files
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for i, path in enumerate(bam_paths):
            f = exe.submit(merge_single_bam,
                           bam_path=path,
                           cell_id_to_cluster=cell_id_to_cluster,
                           output_prefix=f'{output_prefix}{i:06d}',
                           header_dict=header_dict)
            futures[f] = path

        for f in as_completed(futures):
            cluster_read_counts = f.result()
            for k, v in cluster_read_counts.items():
                total_cluster_read_counts[k] += v

    # merge cluster bam files
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for cluster in clusters:
            chunk_paths = list(glob.glob(f'{output_prefix}*_{cluster}.bam'))
            if len(chunk_paths) == 0:
                continue
            merge_cmd = f'samtools merge --no-PG -c -o {output_prefix}_{cluster}.bam ' \
                        f'{output_prefix}*_{cluster}.bam && ' \
                        f'samtools index {output_prefix}_{cluster}.bam'
            f = exe.submit(subprocess.run,
                           merge_cmd,
                           shell=True,
                           check=True)
            futures[f] = chunk_paths

        for f in as_completed(futures):
            chunk_paths = futures[f]
            f.result()
            for path in chunk_paths:
                os.unlink(path)
    return



def mct_config_str(config):
    """Change the dtype of parameters and make a appropriate string"""
    int_parameters = {
        'overlap': 6,
        'r1_left_cut': 10,
        'r1_right_cut': 10,
        'r2_left_cut': 10,
        'r2_right_cut': 10,
        'quality_threshold': 20,
        'length_threshold': 30,
        'total_read_pairs_min': 1,
        'total_read_pairs_max': 6000000,
        'mapq_threshold': 10,
        'num_upstr_bases': 0,
        'num_downstr_bases': 2,
        'compress_level': 5,
        'dna_cov_min_threshold': 3,
        'rna_cov_min_threshold': 3
    }

    float_parameters = {
        'mc_rate_max_threshold': 0.5,
        'mc_rate_min_threshold': 0.9
    }
    bool_parameters = {'unmapped_fastq': False}

    str_parameters = {
        'mode': 'mc',
        'barcode_version': 'required',
        'r1_adapter': 'AGATCGGAAGAGCACACGTCTGAAC',
        'r2_adapter': 'AGATCGGAAGAGCGTCGTGTAGGGA',
        'bismark_reference': 'required',
        'hisat3n_dna_reference': 'required',
        'hisat3n_rna_reference': 'required',
        'hisat3n_repeat_index_type': 'no-repeat',
        'reference_fasta': 'required',
        'star_reference': 'required',
        'gtf_path': 'required',
        'feature_type': 'gene',
        'id_type': 'gene_id',
        'nome_flag_str': 'required'
    }
    if 'hisat3n_dna_reference' in config:
        del str_parameters['bismark_reference']
        del str_parameters['star_reference']

    typed_config = {}
    for k, default in int_parameters.items():
        if k in config:
            typed_config[k] = int(config[k])
        else:
            if default != 'required':
                typed_config[k] = default
            else:
                raise ValueError(f'Required parameter {k} not found in config.')

    for k, default in float_parameters.items():
        if k in config:
            typed_config[k] = float(config[k])
        else:
            if default != 'required':
                typed_config[k] = default
            else:
                raise ValueError(f'Required parameter {k} not found in config.')

    for k, default in bool_parameters.items():
        if k in config:
            v = config[k]
            if v.lower().startswith('t'):
                v = True
            else:
                v = False
            typed_config[k] = v
        else:
            if default != 'required':
                typed_config[k] = default
            else:
                raise ValueError(f'Required parameter {k} not found in config. '
                                 f'You can print the newest mapping config template via "yap default-mapping-config".')
    # judge unmapped_fastq specifically
    unmapped_param_str = '--un' if typed_config['unmapped_fastq'] else ''
    typed_config['unmapped_param_str'] = f"'{unmapped_param_str}'"

    for k, default in str_parameters.items():
        if k in config:
            typed_config[k] = f"'{config[k]}'"
        else:
            if default != 'required':
                typed_config[k] = f"'{default}'"
            else:
                raise ValueError(f'Required parameter {k} not found in config. '
                                 f'You can print the newest mapping config template via "yap default-mapping-config".')

    config_str = ""
    for k, v in typed_config.items():
        config_str += f"{k} = {v}\n"
    return config_str
